#!/usr/bin/env python

#******************************************************************************
#
#  rpnMatchUnitTypes.py
#
#  rpnChilada measurements class and unit conversion
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

from mpmath import mpf


#******************************************************************************
#
#  getWhichUnitType
#
#******************************************************************************

def getWhichUnitType( measurement, unitTypes ):
    for unitType in unitTypes:
        if measurement.isOfUnitType( unitType ):
            return unitType

    return None


#******************************************************************************
#
#  matchUnitTypes
#
#  In addition to measurements, this method also handles RPNDateTime ('datetime')
#  RPNLocation ('location'), and strings (assumed to be locations, and converted
#  to RPNLocation)
#
#******************************************************************************

def matchUnitTypes( args, validUnitTypes ):
    # imported here to avoid circular dependencies
    from rpn.rpnAstronomy import RPNAstronomicalObject
    from rpn.rpnDateTime import RPNDateTime
    from rpn.rpnLocation import getLocation, RPNLocation

    result = { }

    for unitTypeList in validUnitTypes:
        unitTypes = list( unitTypeList )

        if len( args ) != len( unitTypes ):
            raise ValueError( 'argument count mismatch in matchUnitTypes( )' )

        for arg in args:
            if isinstance( arg, mpf ):
                if 'constant' in unitTypes:
                    result[ 'constant' ] = arg
                    continue

                result = { }
                break

            if isinstance( arg, RPNDateTime ):
                if 'datetime' in unitTypes:
                    result[ 'datetime' ] = arg
                    continue

                result = { }
                break

            if isinstance( arg, str ):
                if 'location' in unitTypes:
                    result[ 'location' ] = getLocation( arg )
                    continue

                result = { }
                break

            if isinstance( arg, RPNLocation ):
                if 'location' in unitTypes:
                    result[ 'location' ] = arg
                    continue

                result = { }
                break

            if isinstance( arg, RPNAstronomicalObject ):
                if 'body' in unitTypes:
                    result[ 'body' ] = arg
                    continue

                result = { }
                break

            unitType = getWhichUnitType( arg, unitTypes )
            #print( 'found unit type', unitType )

            if unitType:
                #print( 'setting unitType', unitType )
                result[ unitType ] = arg
            else:
                result = { }
                #print( 'breaking...' )
                #print( )
                break

            unitTypes.remove( unitType )
        else:
            return result

    return None




#!/usr/bin/env python

#******************************************************************************
#
#  rpnCombinatorics.py
#
#  rpnChilada combinatorics operators
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

from mpmath import arange, bell, bernoulli, binomial, e, fac, fadd, fdiv, floor, \
                   fmul, fprod, fsub, fsum, log10, mp, mpmathify, nint, nsum, pi, \
                   power, sqrt, stirling1, stirling2

from rpn.rpnDebug import debugPrint
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnNumberTheory import getNthLinearRecurrence
from rpn.rpnPersistence import cachedFunction
from rpn.rpnPolytope import getNthGeneralizedPolygonalNumber
from rpn.rpnUtils import oneArgFunctionEvaluator, twoArgFunctionEvaluator, validateRealInt


#******************************************************************************
#
#  getNthAperyNumber
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@cachedFunction( 'apery' )
def getNthAperyNumber( n ):
    '''
    http://oeis.org/A005259

    a(n) = sum(k=0..n, C(n,k)^2 * C(n+k,k)^2 )
    '''
    if n < 0:
        raise ValueError( '\'nth_apery\' expects a non-negative argument' )

    precision = int( fmul( n, 1.6 ) )

    if mp.dps < precision:
        mp.dps = precision

    result = 0

    for k in arange( 0, validateRealInt( n ) + 1 ):
        result = fadd( result, fmul( power( binomial( n, k ), 2 ),
                                     power( binomial( fadd( n, k ), k ), 2 ) ) )

    return result


#******************************************************************************
#
#  getNthDelannoyNumber
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@cachedFunction( 'delannoy' )
def getNthDelannoyNumber( n ):
    if validateRealInt( n ) == 1:
        return 3

    if n < 0:
        raise ValueError( '\'nth_delannoy\' expects a non-negative argument' )

    precision = int( fmul( n, 0.8 ) )

    if mp.dps < precision:
        mp.dps = precision

    result = 0

    for k in arange( 0, fadd( validateRealInt( n ), 1 ) ):
        result = fadd( result, fmul( binomial( n, k ), binomial( fadd( n, k ), k ) ) )

    return result


#******************************************************************************
#
#  getNthSchroederNumber
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@cachedFunction( 'schroeder' )
def getNthSchroederNumber( n ):
    if validateRealInt( n ) == 1:
        return 1

    if n < 0:
        raise ValueError( '\'nth_schroeder\' expects a non-negative argument' )

    n = fsub( n, 1 )

    result = 0

    precision = int( fmul( n, 0.8 ) )

    if mp.dps < precision:
        mp.dps = precision

    for k in arange( 0, fadd( n, 1 ) ):
        result = fadd( result, fdiv( fprod( [ power( 2, k ), binomial( n, k ),
                                              binomial( n, fsub( k, 1 ) ) ] ), n ) )

    return result


#******************************************************************************
#
#  getNthMotzkinNumber
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@cachedFunction( 'motzkin' )
def getNthMotzkinNumber( n ):
    '''
    http://oeis.org/A001006

    a(n) = sum((-1)^j*binomial(n+1, j)*binomial(2n-3j, n), j=0..floor(n/3))/(n+1)
    '''
    precision = int( n )

    if mp.dps < precision:
        mp.dps = precision

    result = 0

    for j in arange( 0, floor( fdiv( validateRealInt( n ), 3 ) ) + 1 ):
        result = fadd( result, fprod( [ power( -1, j ), binomial( fadd( n, 1 ), j ),
                                        binomial( fsub( fmul( 2, n ), fmul( 3, j ) ), n ) ] ) )

    return fdiv( result, fadd( n, 1 ) )


#******************************************************************************
#
#  getNthSchroederHipparchusNumber
#
#  https://en.wikipedia.org/wiki/Schr%C3%B6der%E2%80%93Hipparchus_number
#  https://oeis.org/A001003
#
#******************************************************************************

@oneArgFunctionEvaluator( )
@cachedFunction( 'schroeder_hipparchus' )
def getNthSchroederHipparchusNumber( n ):
    if validateRealInt( n ) == 0:
        return 1

    if n < 0:
        raise ValueError( '\'nth_schroeder_hipparchus\' expects a non-negative argument' )

    precision = int( fmul( n, 0.8 ) )

    if mp.dps < precision:
        mp.dps = precision

    result = 0

    for i in arange( n ):
        result = fadd( result, fmul( getNarayanaNumber( n, fadd( i, 1 ) ), power( 2, i ) ) )

    return result


#******************************************************************************
#
#  getNthPellNumber
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthPellNumber( n ):
    '''
    From:  http://oeis.org/A000129

    a( n ) = round( ( 1 + sqrt( 2 ) ) ^ n )
    '''
    if n < 0:
        raise ValueError( '\'nth_pell\' expects a non-negative argument' )

    precision = int( fmul( n, 0.4 ) )

    if mp.dps < precision:
        mp.dps = precision

    return getNthLinearRecurrence( [ 1, 2 ], [ 0, 1 ], fsub( n, 1 ) )


#******************************************************************************
#
#  getPermutations
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getPermutations( n, r ):
    if validateRealInt( r ) > validateRealInt( n ):
        raise ValueError( 'number of elements {0} cannot exceed the size of the set {1}'.format( r, n ) )

    return fdiv( fac( n ), fac( fsub( n, r ) ) )


#******************************************************************************
#
#  getCombinations
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getCombinations( n, r ):
    if validateRealInt( r ) > validateRealInt( n ):
        raise ValueError( 'number of elements {0} cannot exceed the size of the set {1}'.format( r, n ) )

    return fdiv( fac( n ), fmul( fac( fsub( n, r ) ), fac( r ) ) )


#******************************************************************************
#
#  getArrangements
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getArrangements( n ):
    return floor( fmul( fac( n ), e ) )


#******************************************************************************
#
#  getNthSylvesterNumber
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthSylvesterNumber( n ):
    if validateRealInt( n ) == 1:
        return 2
    elif n == 2:
        return 3
    else:
        sylvesters = [ 2, 3 ]

        for _ in arange( 2, n ):
            sylvesters.append( fprod( sylvesters ) + 1 )

    return sylvesters[ -1 ]


#******************************************************************************
#
#  createDeBruijnSequence
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def createDeBruijnSequence( n, k ):
    wordSize = validateRealInt( k )
    symbolCount = validateRealInt( n )

    v = [ 0 for _ in range( wordSize ) ]
    l = 1

    while True:
        if wordSize % l == 0:
            for i in range( 0, l ):
                yield v[ i ]

        for i in range( l, wordSize ):
            v[ i ] = v[ i - l ]

        l = wordSize

        while l > 0 and v[ l - 1 ] >= symbolCount - 1:
            l -= 1

        if l == 0:
            break

        v[ l - 1 ] += 1

def getDeBruijnSequence( n, k ):
    return RPNGenerator( createDeBruijnSequence( n, k ) )


#******************************************************************************
#
#  getCompositions
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getCompositionsGenerator( n, k ):
    value = int( validateRealInt( n ) )
    count = int( validateRealInt( k ) )

    if count < 1:
        raise ValueError( "'compositions' expects a size argument greater than 0'" )

    if count == 1:
        yield [ value ]
    else:
        #result = [ ]

        for i in range( 1, int( ( value - count ) + 2 ) ):
            #result.extend( [ [ nint( i ) ] + comp for comp in getCompositions( n - i, count - 1 ) ] )
            for comp in getCompositionsGenerator( n - i, count - 1 ):
                yield [ nint( i ) ] + comp

@twoArgFunctionEvaluator( )
def getCompositions( n, k ):
    return RPNGenerator( getCompositionsGenerator( n, k ) )


#******************************************************************************
#
#  oldGetPartitionNumber2
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def oldGetPartitionNumber2( n ):
    if validateRealInt( n ) < 0:
        raise ValueError( 'non-negative argument expected' )

    if n in ( 0, 1 ):
        return 1

    total = 0
    sign = 1
    i = 1

    k = getNthGeneralizedPolygonalNumber( i, 5 )

    while n - k >= 0:
        total += sign * getPartitionNumber( fsub( n, k ) )

        i += 1

        if i % 2:
            sign *= -1

        k = getNthGeneralizedPolygonalNumber( i, 5 )

    return total


#public static long partition(long k, long n){
#    long sum = 0;
#    if(k > n) return 0;
#    if(k == n) return 1;
#    sum += partition(k+1, n) + partition(k, n-k);
#    return sum;
#}

def partitionsWithLimit( n, k=None ):
    """Generate all partitions of integer n (>= 0) using integers no
    greater than k (default, None, allows the partition to contain n).

    Each partition is represented as a multiset, i.e. a dictionary
    mapping an integer to the number of copies of that integer in
    the partition.  For example, the partitions of 4 are {4: 1},
    {3: 1, 1: 1}, {2: 2}, {2: 1, 1: 2}, and {1: 4} corresponding to
    [4], [1, 3], [2, 2], [1, 1, 2] and [1, 1, 1, 1], respectively.
    In general, sum(k * v for k, v in a_partition.iteritems()) == n, and
    len(a_partition) is never larger than about sqrt(2*n).

    Note that the _same_ dictionary object is returned each time.
    This is for speed:  generating each partition goes quickly,
    taking constant time independent of n. If you want to build a list
    of returned values then use .copy() to get copies of the returned
    values:

    >>> p_all = []
    >>> for p in partitions(6, 2):
    ...         p_all.append(p.copy())
    ...
    >>> print p_all
    [{2: 3}, {1: 2, 2: 2}, {1: 4, 2: 1}, {1: 6}]

    Reference
    ---------
    Modified from Tim Peter's posting to accomodate a k value:
    http://code.activestate.com/recipes/218332/
    """
    n = int( n )

    if n < 0 or k < 0:
        raise ValueError( '\'partition_limit\' requires a positive count and a positive limit' )

    if n == 0:
        yield [ ]
        return

    if k is None or k > n:
        k = n

    k = int( k )

    quotient, remainder = divmod( n, k )
    itemCounts = { k : quotient }
    keys = [ k ]

    if remainder:
        itemCounts[ remainder ] = 1
        keys.append( remainder )

    result = [ ]

    for key, value in itemCounts.items( ):
        result.extend( [ key ] * value )

    yield result

    while keys != [ 1 ]:
        # Reuse any 1's.
        if keys[ -1 ] == 1:
            del keys[ -1 ]
            reuse = itemCounts.pop( 1 )
        else:
            reuse = 0

        # Let i be the smallest key larger than 1.  Reuse one
        # instance of i.
        i = keys[ -1 ]
        newcount = itemCounts[ i ] = itemCounts[ i ] - 1
        reuse += i

        if newcount == 0:
            del keys[ -1 ], itemCounts[ i ]

        # Break the remainder into pieces of size i - 1.
        i -= 1
        quotient, remainder = divmod( reuse, i )
        itemCounts[ i ] = quotient
        keys.append( i )

        if remainder:
            itemCounts[ remainder ] = 1
            keys.append( remainder )

        result = [ ]

        for key, value in itemCounts.items( ):
            result.extend( [ key ] * value )

        yield result


@twoArgFunctionEvaluator( )
def getPartitionsWithLimit( n, k ):
    if validateRealInt( k ) > validateRealInt( n ):
        k = n

    return RPNGenerator( partitionsWithLimit( n, k ) )


@oneArgFunctionEvaluator( )
@cachedFunction( 'partition' )
def getPartitionNumber( n ):
    '''
    This version is, um, less recursive than the original, which I've kept.
    The strategy is to create a list of the smaller partition numbers we need
    to calculate and then start calling them recursively, starting with the
    smallest.  This will minimize the number of recursions necessary, and in
    combination with caching values, will calculate practically any integer
    partition without the risk of a stack overflow.

    I can't help but think this is still grossly inefficient compared to what's
    possible.  It seems that using this algorithm, calculating any integer
    partition ends up necessitating calculating the integer partitions of
    every integer smaller than the original argument.
    '''
    debugPrint( 'partition', int( n ) )

    if validateRealInt( n ) < 0:
        raise ValueError( 'non-negative argument expected' )

    if n in ( 0, 1 ):
        return 1

    sign = 1
    i = 1
    k = 1

    estimate = log10( fdiv( power( e, fmul( pi, sqrt( fdiv( fmul( 2, n ), 3 ) ) ) ),
                            fprod( [ 4, n, sqrt( 3 ) ] ) ) )
    if mp.dps < estimate + 5:
        mp.dps = estimate + 5

    partitionList = [ ]

    while n - k >= 0:
        partitionList.append( ( fsub( n, k ), sign ) )
        i += 1

        if i % 2:
            sign *= -1

        k = getNthGeneralizedPolygonalNumber( i, 5 )

    partitionList = partitionList[ : : -1 ]

    total = 0

    for partition, sign in partitionList:
        total = fadd( total, fmul( sign, getPartitionNumber( partition ) ) )

    return total


@oneArgFunctionEvaluator( )
@cachedFunction( 'old_partition' )
def oldGetPartitionNumber( n ):
    if n < 0:
        return 0

    if n < 2:
        return 1

    result = mpmathify( 0 )

    for k in arange( 1, n + 1 ):
        #n1 = n - k * ( 3 * k - 1 ) / 2
        sub1 = fsub( n, fdiv( fmul( k, fsub( fmul( 3, k ), 1 ) ), 2 ) )
        #n2 = n - k * ( 3 * k + 1 ) / 2
        sub2 = fsub( n, fdiv( fmul( k, fadd( fmul( 3, k ), 1 ) ), 2 ) )

        result = fadd( result, fmul( power( -1, fadd( k, 1 ) ), \
                       fadd( getPartitionNumber( sub1 ), getPartitionNumber( sub2 ) ) ) )

        if sub1 <= 0:
            break

    #old = NOT_QUITE_AS_OLDgetPartitionNumber( n )
    #
    #if ( old != result ):
    #    raise ValueError( "It's broke." )

    return result


#******************************************************************************
#
#  createIntegerPartitions
#
#  https://code.activestate.com/recipes/218332-generator-for-integer-partitions/
#
#  http://jeromekelleher.net/generating-integer-partitions.html provides a
#  similar version of this algorithm.
#
#******************************************************************************

def createIntegerPartitions( n ):
    """Generate partitions of n as ordered lists in ascending
    lexicographical order.

    This highly efficient routine is based on the delightful
    work of Kelleher and O'Sullivan.

    Examples
    ========

    >>> for i in aP(6): i
    ...
    [1, 1, 1, 1, 1, 1]
    [1, 1, 1, 1, 2]
    [1, 1, 1, 3]
    [1, 1, 2, 2]
    [1, 1, 4]
    [1, 2, 3]
    [1, 5]
    [2, 2, 2]
    [2, 4]
    [3, 3]
    [6]

    >>> for i in aP(0): i
    ...
    []

    References
    ==========

    .. [1] Generating Integer Partitions, [online],
        Available: http://jeromekelleher.net/generating-integer-partitions.html
    .. [2] Jerome Kelleher and Barry O'Sullivan, "Generating All
        Partitions: A Comparison Of Two Encodings", [online],
        Available: http://arxiv.org/pdf/0909.2331v2.pdf

    """
    # The list `a`'s leading elements contain the partition in which
    # y is the biggest element and x is either the same as y or the
    # 2nd largest element; v and w are adjacent element indices
    # to which x and y are being assigned, respectively.
    a = [ 1 ] * n
    y = -1
    v = n

    while v > 0:
        v -= 1
        x = a[ v ] + 1

        while y >= 2 * x:
            a[ v ] = x
            y -= x
            v += 1

        w = v + 1

        while x <= y:
            a[ v ] = x
            a[ w ] = y
            yield a[ : w + 1 ]
            x += 1
            y -= 1

        a[ v ] = x + y
        y = a[ v ] - 1
        yield a[ : w ]


#******************************************************************************
#
#  getIntegerPartitions
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getIntegerPartitions( n ):
    return RPNGenerator( createIntegerPartitions( int( n ) ) )


#******************************************************************************
#
#  getNthMultifactorial
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getNthMultifactorial( n, k ):
    return fprod( arange( validateRealInt( n ), 0, -( validateRealInt( k ) ) ) )


#******************************************************************************
#
#  getMultinomial
#
#******************************************************************************

def getMultinomial( args ):
    numerator = fac( fsum( args ) )

    denominator = 1

    for arg in args:
        denominator = fmul( denominator, fac( arg ) )

    return fdiv( numerator, denominator )


#******************************************************************************
#
#  getLahNumber
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getLahNumber( n, k ):
    return fmul( power( -1, n ), fdiv( fmul( binomial( validateRealInt( n ), validateRealInt( k ) ),
                                             fac( fsub( n, 1 ) ) ), fac( fsub( k, 1 ) ) ) )


#******************************************************************************
#
#  getNarayanaNumber
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getNarayanaNumber( n, k ):
    return fdiv( fmul( binomial( n, k ), binomial( n, fsub( k, 1 ) ) ), n )

@twoArgFunctionEvaluator( )
def getNarayanaNumberOperator( n, k ):
    return getNarayanaNumber( n, k )


#******************************************************************************
#
#  getNthCatalanNumber
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthCatalanNumber( n ):
    return fdiv( binomial( fmul( 2, validateRealInt( n ) ), n ), fadd( n, 1 ) )


#******************************************************************************
#
#  getNthMenageNumber
#
#  https://oeis.org/A000179
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getNthMenageNumber( n ):
    if n < 0:
        raise ValueError( '\'menage\' requires a non-negative argument' )

    if n == 1:
        return -1
    elif n == 2:
        return 0
    elif n in [ 0, 3 ]:
        return 1
    else:
        return nsum( lambda k: fdiv( fprod( [ power( -1, k ), fmul( 2, n ), binomial( fsub( fmul( 2, n ), k ), k ),
                                              fac( fsub( n, k ) ) ] ), fsub( fmul( 2, n ), k ) ), [ 0, n ] )

@twoArgFunctionEvaluator( )
def getBellPolynomial( n, k ):
    return bell( n, k )

@twoArgFunctionEvaluator( )
def getBinomial( n, k ):
    return binomial( n, k )

@oneArgFunctionEvaluator( )
def getNthBell( n ):
    return bell( n )

@oneArgFunctionEvaluator( )
def getNthBernoulli( n ):
    return bernoulli( n )


#******************************************************************************
#
#  countFrobenius
#
#  https://math.stackexchange.com/questions/176363/keep-getting-generating- \
#      function-wrong-making-change-for-a-dollar/176397#176397
#
#  Here's another way from Sloane that doesn't require so much memory:
#
#  a(n) = (1/n)*Sum_{k=1..n} A008472(k)*a(n-k).
#
#******************************************************************************

def countFrobenius( denominations, target ):
    target = int( target )
    data = [ 0 ] * ( target + 1 )
    data[ 0 ] = 1

    for k in denominations:
        k = int( k )

        for i in range( 0, target - k + 1 ):
            data[ i + k ] += data[ i ]

    return data[ target ]


#******************************************************************************
#
#  getStirling1Number
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getStirling1Number( n, k ):
    return stirling1( n, k )


#******************************************************************************
#
#  getStirling2Number
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getStirling2Number( n, k ):
    return stirling2( n, k )


#  http://crowsandcats.blogspot.com/2013/04/approximate-restricted-integer.html

#Recursive restricted block partition. Each call reduces the sum, n,
# and m, the number of parts.
#def blockpart(n, m, k):
#
#    #Is the number of parts only 1?
#    if(m == 1):
#        #Trivial: to partition n into one part, we need to be able to use  n
#        if(n < k):
#            return 1
#        else:
#            return 0
#    #what is the highest possible sum we may have
#    #thatbb will still be a valid partition
#    smax = m * (k - 1)
#    #This is the lowest term we can add to the partition (subtract from sum)
#    start = n - k + 1
#
#    #If the lowest is too high, then we aren't able to get a valid partition
#    if( start > smax):
#        return 0
#
#    paths = 0
#
#    #add up all possible other partitions after subtracting choices
#    #to place in the current part (m)
#    for i in range(max(0,start), 1 + min(smax, n)):
#        paths += blockpart(i, m - 1, k)
#
#    return paths





#!/usr/bin/env python

#******************************************************************************
#
#  preparePrimeData.py
#
#  rpnChilada prime number data file compiler
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  *** NOTE:  Don't run this file directly.  Use ../preparePrimeData.py.
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import os
import time

from rpn.rpnPersistence import createPrimeCache, deleteCache, saveToCache
from rpn.rpnUtils import getSourcePath
from rpn.rpnVersion import PROGRAM_VERSION_STRING, COPYRIGHT_MESSAGE

if not hasattr( time, 'time_ns' ):
    from rpn.rpnNanoseconds import time_ns
else:
    from time import time_ns


#******************************************************************************
#
#  preparePrimeData
#
#******************************************************************************

def preparePrimeData( baseName ):
    print( 'processing ' + baseName + '...' )

    inputFileName = getSourcePath( ) + os.sep + baseName + '.txt'

    deleteCache( baseName )
    db, cursor = createPrimeCache( baseName )

    with open( inputFileName, 'r' ) as input:
        for line in input:
            try:
                key, value = line.split( )
                saveToCache( db, cursor, key, value, commit=False )
            except:
                print( 'key', key )
                print( 'value', value )
                print( 'parsing error in file ' + inputFileName + ': \'' + line + '\'' )
                break

    db.commit( )
    db.close( )


#******************************************************************************
#
#  main
#
#******************************************************************************

def main( ):
    print( 'preparePrimeData' + PROGRAM_VERSION_STRING + ' - rpnChilada prime number data file converter' )
    print( COPYRIGHT_MESSAGE )
    print( )

    startTime = time_ns( )

    preparePrimeData( 'balanced_primes' )
    preparePrimeData( 'cousin_primes' )
    preparePrimeData( 'double_balanced_primes' )
    preparePrimeData( 'huge_primes' )
    preparePrimeData( 'isolated_primes' )
    preparePrimeData( 'large_primes' )
    preparePrimeData( 'octy_primes' )
    preparePrimeData( 'quadruple_balanced_primes' )
    preparePrimeData( 'quad_primes' )
    preparePrimeData( 'quint_primes' )
    preparePrimeData( 'sext_primes' )
    preparePrimeData( 'sexy_primes' )
    preparePrimeData( 'sexy_quadruplets' )
    preparePrimeData( 'sexy_triplets' )
    preparePrimeData( 'small_primes' )
    preparePrimeData( 'sophie_primes' )
    preparePrimeData( 'triplet_primes' )
    preparePrimeData( 'triple_balanced_primes' )
    preparePrimeData( 'twin_primes' )

    print( )
    print( 'Prime number data prepared.  Time elapsed:  {:.3f} seconds'.format( ( time_ns( ) - startTime ) / 1000000000 ) )



#******************************************************************************
#
#  __main__
#
#******************************************************************************

if __name__ == '__main__':
    main( )




#!/usr/bin/env python

#******************************************************************************
#
#  rpnPersistence.py
#
#  rpnChilada persistence functions
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import bz2
import configparser
import contextlib
import functools
import os
import pickle
import sqlite3

from collections import MutableMapping
from functools import lru_cache
from shutil import copyfile

from rpn.rpnDebug import debugPrint
from rpn.rpnGenerator import RPNGenerator
from rpn.rpnKeyboard import DelayedKeyboardInterrupt
from rpn.rpnUtils import getUserDataPath, oneArgFunctionEvaluator
from rpn.rpnVersion import PROGRAM_VERSION, PROGRAM_NAME

import rpn.rpnGlobals as g


#******************************************************************************
#
#  loadFactorCache
#
#******************************************************************************

@lru_cache( 1 )
def loadFactorCache( ):
    g.factorCache = PersistentDict( getUserDataPath( ) + os.sep + 'factors.cache' )


#******************************************************************************
#
#  loadUnitNameData
#
#******************************************************************************

def loadUnitNameData( ):
    try:
        with contextlib.closing( bz2.BZ2File( getUserDataPath( ) + os.sep + 'unit_names.pckl.bz2', 'rb' ) ) \
                as pickleFile:
            unitsVersion = pickle.load( pickleFile )
            g.unitOperatorNames = pickle.load( pickleFile )
            g.constantOperatorNames = pickle.load( pickleFile )
            g.aliases.update( pickle.load( pickleFile ) )
    except IOError:
        print( 'rpn:  Unable to load unit names.  Run "makeUnits" to generate the unit data files.' )
        return False

    if unitsVersion != PROGRAM_VERSION:
        print( 'rpn:  units data file version mismatch.  Run "makeUnits" to generate the unit data files.' )

    return True


#******************************************************************************
#
#  loadUnitConversionMatrix
#
#******************************************************************************

def loadUnitConversionMatrix( ):
    try:
        with contextlib.closing( bz2.BZ2File( getUserDataPath( ) + os.sep + 'unit_conversions.pckl.bz2', 'rb' ) ) \
                as pickleFile:
            g.unitConversionMatrix.update( pickle.load( pickleFile ) )
    except FileNotFoundError:
        print( 'rpn:  Unable to load unit conversion data.  Run "makeUnits" to generate the unit data files.' )


#******************************************************************************
#
#  loadUnitData
#
#******************************************************************************

def loadUnitData( ):
    try:
        with contextlib.closing( bz2.BZ2File( getUserDataPath( ) + os.sep + 'units.pckl.bz2', 'rb' ) ) \
                as pickleFile:
            unitsVersion = pickle.load( pickleFile )
            g.basicUnitTypes.update( pickle.load( pickleFile ) )
            g.unitOperators.update( pickle.load( pickleFile ) )
            g.constantOperators.update( pickle.load( pickleFile ) )
    except IOError:
        print( 'rpn:  Unable to load unit info data.  Run "makeUnits" to generate the unit data files.' )
        return False

    if unitsVersion != PROGRAM_VERSION:
        print( 'rpn:  units data file version mismatch.  Run "makeUnits" to generate the unit data files.' )
        return False

    return True


#******************************************************************************
#
#  loadHelpData
#
#******************************************************************************

def loadHelpData( ):
    if g.helpLoaded:
        return True

    try:
        with contextlib.closing( bz2.BZ2File( getUserDataPath( ) + os.sep + 'help.pckl.bz2', 'rb' ) ) as pickleFile:
            g.helpVersion = pickle.load( pickleFile )
            g.helpTopics = pickle.load( pickleFile )
            g.operatorHelp = pickle.load( pickleFile )
    except FileNotFoundError:
        raise ValueError( 'rpn:  Unable to load help.  Run "makeHelp" to generate the help data files.' )

    try:
        with contextlib.closing( bz2.BZ2File( getUserDataPath( ) + os.sep + 'unit_help.pckl.bz2', 'rb' ) ) \
                as pickleFile:
            g.unitTypeDict = pickle.load( pickleFile )
    except FileNotFoundError:
        raise ValueError( 'rpn:  Unable to load unit help data.  Run "makeHelp" to generate the help data files.' )

    g.operatorCategories = set( g.operatorHelp[ key ][ 0 ] for key in g.operatorHelp )

    g.helpLoaded = True

    return True


#******************************************************************************
#
#  loadResult
#
#******************************************************************************

def loadResult( ):
    try:
        fileName = getUserDataPath( ) + os.sep + 'result.pckl.bz2'

        with contextlib.closing( bz2.BZ2File( fileName, 'rb' ) ) as pickleFile:
            result = pickle.load( pickleFile )
    except FileNotFoundError:
        result = 0

    return result


#******************************************************************************
#
#  saveResult
#
#******************************************************************************

def saveResult( result ):
    fileName = getUserDataPath( ) + os.sep + 'result.pckl.bz2'

    # TODO:  handle RPNGenerator and RPNMeasurement
    try:
        with DelayedKeyboardInterrupt( ):
            with contextlib.closing( bz2.BZ2File( fileName, 'wb' ) ) as pickleFile:
                pickle.dump( result, pickleFile )
    except:
        pass #print( 'error:  failed to save result' )


#******************************************************************************
#
#  loadUserConstants
#
#******************************************************************************

def loadConstants( ):
    try:
        with contextlib.closing( bz2.BZ2File( getUserDataPath( ) + os.sep + 'constants.pckl.bz2', 'rb' ) ) \
                as pickleFile:
            constants = pickle.load( pickleFile )
    except FileNotFoundError:
        constants = { }

    return constants


#******************************************************************************
#
#  saveConstants
#
#******************************************************************************

def saveConstants( constants ):
    with DelayedKeyboardInterrupt( ):
        with contextlib.closing( bz2.BZ2File( getUserDataPath( ) + os.sep + 'constants.pckl.bz2', 'wb' ) ) \
                as pickleFile:
            pickle.dump( constants, pickleFile )


#******************************************************************************
#
#  getCacheFileName
#
#******************************************************************************

@lru_cache( 10 )
def getCacheFileName( name ):
    return getUserDataPath( ) + os.sep + name + '.cache'


#******************************************************************************
#
#  getPrimeCacheFileName
#
#******************************************************************************

@lru_cache( 10 )
def getPrimeCacheFileName( name ):
    return getUserDataPath( ) + os.sep + name + '.cache'


#******************************************************************************
#
#  deleteCache
#
#******************************************************************************

def deleteCache( name ):
    if doesCacheExist( name ):
        os.remove( getCacheFileName( name ) )


#******************************************************************************
#
#  saveToCache
#
#******************************************************************************

def saveToCache( cacheDB, cursor, key, value, commit=True ):
    cursor.execute( '''INSERT INTO cache( id, value ) VALUES( ?, ? )''', ( key, value ) )

    if commit:
        cacheDB.commit( )


#******************************************************************************
#
#  doesCacheExist
#
#******************************************************************************

def doesCacheExist( name ):
    return os.path.isfile( getCacheFileName( name ) )


#******************************************************************************
#
#  createPrimeCache
#
#******************************************************************************

def createPrimeCache( name ):
    cacheDB = sqlite3.connect( getCacheFileName( name ) )

    cursor = cacheDB.cursor( )
    cursor.execute( '''CREATE TABLE cache( id INTEGER PRIMARY KEY, value INTEGER )''' )
    cacheDB.commit( )

    return cacheDB, cursor


#******************************************************************************
#
#  openPrimeCache
#
#******************************************************************************

def openPrimeCache( name ):
    if name not in g.cursors:
        try:
            g.databases[ name ] = sqlite3.connect( getPrimeCacheFileName( name ) )
            g.cursors[ name ] = g.databases[ name ].cursor( )
        except:
            raise ValueError( 'prime number table ' + name +
                              ' cannot be found.  Run "preparePrimeData" to create the prime data.' )

    return g.cursors[ name ]



#******************************************************************************
#
#  dumpPrimeCache
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def dumpPrimeCache( name ):
    if name not in g.cursors:
        if not doesCacheExist( name ):
            raise ValueError( 'cache \'' + name + '\' does not exist.' )

        openPrimeCache( name )

    rows = g.cursors[ name ].execute( 'SELECT id, value FROM cache ORDER BY id' ).fetchall( )

    rows.sort( key=lambda x: x[ 0 ] )

    for row in rows:
        print( '{:13} {}'.format( row[ 0 ], row[ 1 ] ) )

    return len( rows )


#******************************************************************************
#
#  class PersistentDict
#
#  http://stackoverflow.com/questions/9320463/persistent-memoization-in-python
#
#******************************************************************************

class PersistentDict( MutableMapping ):
    def __init__( self, dbpath, iterable=None, **kwargs ):
        self.dbpath = dbpath

        with self.getConnection( ) as connection:
            cursor = connection.cursor( )
            cursor.execute( 'create table if not exists memo '
                            '(key blob primary key not null, value blob not null)' )

        if iterable is not None:
            self.update( iterable )

        self.update( kwargs )

    def encode( self, obj ):
        return pickle.dumps( obj )

    def decode( self, blob ):
        return pickle.loads( blob )

    def getConnection( self ):
        return sqlite3.connect( self.dbpath )

    def  __getitem__( self, key ):
        key = self.encode( key )

        with self.getConnection( ) as connection:
            cursor = connection.cursor( )
            cursor.execute( 'select value from memo where key=?', ( key, ) )

            value = cursor.fetchone( )

        if value is None:
            raise KeyError( key )

        return self.decode( value[ 0 ] )

    def __setitem__( self, key, value ):
        key = self.encode( key )
        value = self.encode( value )

        with self.getConnection( ) as connection:
            cursor = connection.cursor( )
            cursor.execute( 'insert or replace into memo values (?, ?)', ( key, value ) )

    def __delitem__( self, key ):
        key = self.encode( key )

        with self.getConnection( ) as connection:
            cursor = connection.cursor( )

            cursor.execute( 'select count(*) from memo where key=?', ( key, ) )

            if cursor.fetchone( )[ 0 ] == 0:
                raise KeyError( key )

            cursor.execute( 'delete from memo where key=?', ( key, ) )

    def __iter__( self ):
        with self.getConnection( ) as connection:
            cursor = connection.cursor( )
            cursor.execute( 'select key from memo' )

            records = cursor.fetchall( )

        for record in records:
            yield self.decode( record[ 0 ] )

    def __len__( self ):
        with self.getConnection( ) as connection:
            cursor = connection.cursor( )
            cursor.execute( 'select count(*) from memo' )

            return cursor.fetchone( )[ 0 ]


#******************************************************************************
#
#  getUserVariablesFileName
#
#******************************************************************************

@lru_cache( 1 )
def getUserVariablesFileName( ):
    return getUserDataPath( ) + os.sep + PROGRAM_NAME + '_user_variables.cfg'


#******************************************************************************
#
#  getUserFunctionsFileName
#
#******************************************************************************

@lru_cache( 1 )
def getUserFunctionsFileName( ):
    return getUserDataPath( ) + os.sep + PROGRAM_NAME + '_user_functions.cfg'


#******************************************************************************
#
#  getUserConfigurationFileName
#
#******************************************************************************

@lru_cache( 1 )
def getUserConfigurationFileName( ):
    return getUserDataPath( ) + os.sep + PROGRAM_NAME + '_user_config.cfg'


#******************************************************************************
#
#  loadUserVariablesFile
#
#******************************************************************************

def loadUserVariablesFile( ):
    config = configparser.ConfigParser( )
    config.read( getUserVariablesFileName( ) )

    try:
        items = config.items( 'User Variables' )
    except:
        return

    for item in items:
        g.userVariables[ item[ 0 ] ] = item[ 1 ]


#******************************************************************************
#
#  saveUserVariablesFile
#
#******************************************************************************

def saveUserVariablesFile( ):
    config = configparser.ConfigParser( )

    config[ 'User Variables' ] = { }

    for key in g.userVariables:
        config[ 'User Variables' ][ key ] = str( g.userVariables[ key ] )

    if os.path.isfile( getUserVariablesFileName( ) ):
        copyfile( getUserVariablesFileName( ), getUserVariablesFileName( ) + '.backup' )

    with open( getUserVariablesFileName( ), 'w' ) as userVariablesFile:
        config.write( userVariablesFile )


#******************************************************************************
#
#  openFunctionCache
#
#******************************************************************************

def openFunctionCache( name ):
    if name in g.functionCaches:
        return g.functionCaches[ name ]
    else:
        debugPrint( 'opening', name, 'function cache database' )
        g.functionCaches[ name ] = PersistentDict( getCacheFileName( name ) )
        return g.functionCaches[ name ]


#******************************************************************************
#
#  deleteFromFunctionCache
#
#******************************************************************************

#def deleteFromFunctionCache( name, key ):
#    pass
#    #goobles
#    #if name in g.functionCaches:
#    #    return g.functionCaches[ name ]
#    #else:
#    #    debugPrint( 'opening', name, 'function cache database' )
#    #    g.functionCaches[ name ] = PersistentDict( getCacheFileName( name ) )
#    #    return g.functionCaches[ name ]


#******************************************************************************
#
#  dumpFunctionCache
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def dumpFunctionCache( name ):
    if not doesCacheExist( name ):
        raise ValueError( 'cache \'' + name + '\' does not exist.' )

    cache = openFunctionCache( name )

    keys = sorted( cache.keys( ) )

    for key in keys:
        print( key, cache[ key ] )

    return len( cache )


#******************************************************************************
#
#  cachedFunction
#
#  This is a decorator for any function that wishes to cache its calculations
#  to disk.
#
#******************************************************************************

def cachedFunction( name, overrideIgnore=False ):
    def namedCachedFunction( func ):
        @functools.wraps( func )

        def cacheResults( *args, **kwargs ):
            cache = openFunctionCache( name )

            if not g.ignoreCache or overrideIgnore:
                if ( args, kwargs ) in cache:
                    return cache[ ( args, kwargs ) ]

            result = func( *args, **kwargs )

            if isinstance( result, RPNGenerator ):
                result = list( result )

            if not g.ignoreCache or overrideIgnore:
                cache[ ( args, kwargs ) ] = result

            return result

        return cacheResults

    return namedCachedFunction


#******************************************************************************
#
#  cachedOEISFunction
#
#  This is a modified version of cachedFunction( ) that will ignore (and
#  overwrite) the cached result if it equals 0.  This prevents a failed
#  HTTP connection from polluting the OEIS cache with invalid data.
#
#******************************************************************************

def cachedOEISFunction( name, overrideIgnore=False ):
    def namedCachedFunction( func ):
        @functools.wraps( func )

        def cacheResults( *args, **kwargs ):
            cache = openFunctionCache( name )

            if not g.ignoreCache or overrideIgnore:
                if ( args, kwargs ) in cache:
                    result = cache[ ( args, kwargs ) ]

                    if result != 0:
                        return result

            result = func( *args, **kwargs )

            if not g.ignoreCache or overrideIgnore:
                cache[ ( args, kwargs ) ] = result

            return result

        return cacheResults

    return namedCachedFunction


#******************************************************************************
#
#  loadUserConfigurationFile
#
#******************************************************************************

def loadUserConfigurationFile( ):
    config = configparser.ConfigParser( )
    config.read( getUserConfigurationFileName( ) )

    try:
        items = config.items( 'User Configuration' )
    except:
        return

    for item in items:
        g.userConfiguration[ item[ 0 ] ] = item[ 1 ]


#******************************************************************************
#
#  saveUserConfigurationFile
#
#******************************************************************************

def saveUserConfigurationFile( ):
    config = configparser.ConfigParser( )

    config[ 'User Configuration' ] = { }

    for key in g.userConfiguration:
        config[ 'User Configuration' ][ key ] = g.userConfiguration[ key ]

    if os.path.isfile( getUserConfigurationFileName( ) ):
        copyfile( getUserConfigurationFileName( ), getUserConfigurationFileName( ) + '.backup' )

    with open( getUserConfigurationFileName( ), 'w' ) as userConfigurationFile:
        config.write( userConfigurationFile )





#!/usr/bin/env python

#******************************************************************************
#
#  rpnEclipse.py
#
#  rpnChilada eclipse calculation operators
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#  adapted from:
#
#  Javascript Lunar Eclipse Explorer
#  Version 1.1 by Chris O'Byrne and Fred Espenak - 2014 Mar 18.
#
#  This code is being released under the terms of the GNU General Public
#  License (http://www.gnu.org/copyleft/gpl.html) with the request
#  that if you do improve on it or use it in your own site,
#  please let us know at
#  chris@obyrne.com  and  fred espenak (www.eclipsewise.com)    Thanks!
#
#  http://www.eclipsewise.com/lunar/JLEX/JLEX-index.html
#
#******************************************************************************

import ephem

//
// Observer constants -
// (0) North Latitude (radians)
// (1) West Longitude (radians)
// (2) Altitude (metres)
// (3) West time zone (hours)
// (4) index into the elements array for the eclipse in question
// (5) maximum eclipse type
//
var obsvconst = new Array();

//
// Eclipse circumstances
//  (0) Event type (p1=-3, u1=-2, u2=-1, mid=0, u3=1, u4=2, p4=3)
//  (1) t
//  (2) hour angle
//  (3) declination
//  (4) alt
//  (5) visibility
//      (0 = above horizon, 1 = no event, 2 = below horizon)
//

var month = new Array("Jan","Feb","Mar","Apr","May","Jun","Jul",
                      "Aug","Sep","Oct","Nov","Dec");

var p1 = new Array();
var u1 = new Array();
var u2 = new Array();
var mid = new Array();
var u3 = new Array();
var u4 = new Array();
var p4 = new Array();


var currenttimeperiod = "";
var loadedtimeperiods = new Array();

//
// Populate the circumstances array
// entry condition - circumstances[1] must contain the correct value
function populatecircumstances(elements, circumstances) {
  var index, t, ra, dec, h;

  index = obsvconst[4]
  t = circumstances[1]
  ra = elements[18+index] * t + elements[17+index]
  ra = ra * t + elements[16+index]
  dec = elements[21+index] * t + elements[20+index]
  dec = dec * t + elements[19+index]
  dec = dec * Math.PI / 180.0
  circumstances[3] = dec
  h = 15.0*(elements[6+index] + (t - elements[2+index]/3600.0)*1.00273791) - ra
  h = h * Math.PI / 180.0 - obsvconst[1]
  circumstances[2] = h
  circumstances[4] = Math.asin(Math.sin(obsvconst[0]) * Math.sin(dec) + Math.cos(obsvconst[0]) * Math.cos(dec) * Math.cos(h))
  circumstances[4] -= Math.asin(Math.sin(elements[7+index]*Math.PI/180.0) * Math.cos(circumstances[4]))
  if (circumstances[4] * 180.0 / Math.PI < elements[8+index] - 0.5667) {
    circumstances[5] = 2
  } else if (circumstances[4] < 0.0) {
    circumstances[4] = 0.0
    circumstances[5] = 0
  } else {
    circumstances[5] = 0
  }
}

//
// Populate the p1, u1, u2, mid, u3, u4 and p4 arrays
function getall(elements) {
  var pattern, index

  index = obsvconst[4]
  p1[1] = elements[index+9]
  populatecircumstances(elements,p1)
  mid[1] = elements[index+12]
  populatecircumstances(elements,mid)
  p4[1] = elements[index+15]
  populatecircumstances(elements,p4)
  if (elements[index+5] < 3) {
    u1[1] = elements[index+10]
    populatecircumstances(elements,u1)
    u4[1] = elements[index+14]
    populatecircumstances(elements,u4)
    if (elements[index+5] < 2) {
      u2[1] = elements[index+11]
      u3[1] = elements[index+13]
      populatecircumstances(elements,u2)
      populatecircumstances(elements,u3)
    } else {
      u2[5] = 1
      u3[5] = 1
    }
  } else {
    u1[5] = 1
    u2[5] = 1
    u3[5] = 1
    u4[5] = 1
  }
  if ((p1[5] != 0) && (u1[5] != 0) && (u2[5] != 0) && (mid[5] != 0) && (u3[5] != 0) && (u4[5] != 0) && (p4[5] != 0)) {
    mid[5] = 1
  }
}

//
// Read the data that's in the form, and populate the obsvconst array
function readform() {
  var tmp

  // Make sure that we have something to parse from the form
  if (document.eclipseform.latd.value == "") document.eclipseform.latd.value="0"
  if (document.eclipseform.latm.value == "") document.eclipseform.latm.value="0"
  if (document.eclipseform.lats.value == "") document.eclipseform.lats.value="0"
  if (document.eclipseform.lond.value == "") document.eclipseform.lond.value="0"
  if (document.eclipseform.lonm.value == "") document.eclipseform.lonm.value="0"
  if (document.eclipseform.lons.value == "") document.eclipseform.lons.value="0"
  if (document.eclipseform.alt.value == "") document.eclipseform.alt.value="0"

  // Write back to the form what we are parsing
  document.eclipseform.latd.value=Math.abs(parseFloat(document.eclipseform.latd.value))
  document.eclipseform.latm.value=Math.abs(parseFloat(document.eclipseform.latm.value))
  document.eclipseform.lats.value=Math.abs(parseFloat(document.eclipseform.lats.value))
  document.eclipseform.lond.value=Math.abs(parseFloat(document.eclipseform.lond.value))
  document.eclipseform.lonm.value=Math.abs(parseFloat(document.eclipseform.lonm.value))
  document.eclipseform.lons.value=Math.abs(parseFloat(document.eclipseform.lons.value))
  document.eclipseform.alt.value=Math.abs(parseFloat(document.eclipseform.alt.value))

  // Get the latitude
  obsvconst[0]=parseFloat(document.eclipseform.latd.value)+parseFloat(document.eclipseform.latm.value)/60.0+parseFloat(document.eclipseform.lats.value)/3600.0
  obsvconst[0]=obsvconst[0]*parseFloat(document.eclipseform.latx.options[document.eclipseform.latx.selectedIndex].value)
  obsvconst[0]=obsvconst[0]*Math.PI/180.0

  // Get the longitude
  obsvconst[1]=parseFloat(document.eclipseform.lond.value)+parseFloat(document.eclipseform.lonm.value)/60.0+parseFloat(document.eclipseform.lons.value)/3600.0
  obsvconst[1]=obsvconst[1]*parseFloat(document.eclipseform.lonx.options[document.eclipseform.lonx.selectedIndex].value)
  obsvconst[1]=obsvconst[1]*Math.PI/180.0

  // Get the altitude
  obsvconst[2]=parseFloat(document.eclipseform.alt.value)

  // Get the time zone
  obsvconst[3]=parseFloat(document.eclipseform.tzm.options[document.eclipseform.tzm.selectedIndex].value)
  obsvconst[3]=parseFloat(document.eclipseform.tzh.options[document.eclipseform.tzh.selectedIndex].value) + obsvconst[3]/60.0
  obsvconst[3]=parseFloat(document.eclipseform.tzx.options[document.eclipseform.tzx.selectedIndex].value) * obsvconst[3]

  obsvconst[4] = 0
  for (tmp = 0 ; tmp < 3 ; tmp++) {
    if (document.eclipseform.ecltype[tmp].checked) {
      obsvconst[5] = document.eclipseform.ecltype[tmp].value
    }
  }

}

//
// Get the local date of an event
function getdate(elements,circumstances) {
  var t, ans, jd, a, b, c, d, e, index

  index = obsvconst[4]
  // Calculate the JD for noon (TDT) the day before the day that contains T0
  jd = Math.floor(elements[index] - (elements[1+index]/24.0))
  // Calculate the local time (ie the offset in hours since midnight TDT on the day containing T0).
  t = circumstances[1] + elements[1+index] - obsvconst[3] - (elements[2+index] - 30.0) / 3600.0
  if (t < 0.0) {
    jd--;
  }
  if (t >= 24.0) {
    jd++;
  }
  if (jd >= 2299160.0) {
    a = Math.floor((jd - 1867216.25) / 36524.25)
    a = jd + 1 + a - Math.floor(a/4);
  } else {
    a = jd;
  }
  b = a + 1525.0
  c = Math.floor((b-122.1)/365.25)
  d = Math.floor(365.25*c)
  e = Math.floor((b - d) / 30.6001)
  d = b - d - Math.floor(30.6001*e)
  if (e < 13.5) {
    e = e - 1
  } else {
    e = e - 13
  }
  if (e > 2.5) {
    ans = c - 4716 + "-"
  } else {
    ans = c - 4715 + "-"
  }
  ans += month[e-1] + "-"
  if (d < 10) {
    ans = ans + "0"
  }
  ans = ans + d
  return ans
}

//
// Get the local time of an event
function gettime(elements,circumstances) {
  var t, ans, index, html, ital

  ans = ""
  index = obsvconst[4]
  t = circumstances[1] + elements[1+index] - obsvconst[3] - (elements[2+index] - 30.0) / 3600.0
  if (t < 0.0) {
    t = t + 24.0
  }
  if (t >= 24.0) {
    t = t - 24.0
  }
  if (t < 10.0) {
    ans = ans + "0"
  }
  ans = ans + Math.floor(t) + ":"
  t = (t * 60.0) - 60.0 * Math.floor(t)
  if (t < 10.0) {
    ans = ans + "0"
  }
  ans = ans + Math.floor(t)
  if (circumstances[5] == 2) {
    html = document.createElement("font");
    html.setAttribute("color","#808080");
    ital = document.createElement("i");
    ital.appendChild(document.createTextNode(ans));
    html.appendChild(ital);
    return html;
  } else {
    return document.createTextNode(ans);
  }
}

//
// Get the altitude
function getalt(circumstances) {
  var t, ans, html, ital

  t = circumstances[4] * 180.0 / Math.PI
  t = Math.floor(t+0.5)
  if (t < 0.0) {
    ans = "-"
    t = -t
  } else {
    ans = "+"
  }
  if (t < 10.0) {
    ans = ans + "0"
  }
  ans = ans + t
  if (circumstances[5] == 2) {
    html = document.createElement("font");
    html.setAttribute("color","#808080");
    ital = document.createElement("i");
    ital.appendChild(document.createTextNode(ans));
    html.appendChild(ital);
    return html;
  } else {
    return document.createTextNode(ans);
  }
}


function clearoldresults() {
  results = document.getElementById("el_results");
  resultsTable = document.getElementById("el_locationtable");
  if (resultsTable != null) results.removeChild(resultsTable);
  resultsTable = document.getElementById("el_resultstable");
  if (resultsTable != null) results.removeChild(resultsTable);
}

// CALCULATE!
function calculatefor(el) {
  readform()
  clearoldresults();
  results = document.getElementById("el_results");
  p = document.createElement("p");
  p.setAttribute("id","el_locationtable");
  b = document.createElement("h2");
  b.setAttribute("align","center");
  b.appendChild(document.createTextNode("Lunar Eclipses from "+el[0]));
  p.appendChild(b);
  b = document.createElement("h2");
  b.setAttribute("align","center");
  b.appendChild(document.createTextNode(document.eclipseform.loc_name.value));
  p.appendChild(b);
  resultsTable = document.createElement("table");
  resultsTable.setAttribute("border","0");
  tbody = document.createElement("tbody");
  row = document.createElement("tr");
  td = document.createElement("td");
  td.setAttribute("align","right");
  td.setAttribute("nowrap","");
  td.appendChild(document.createTextNode("Latitude: "));
  row.appendChild(td);
  td = document.createElement("td");
  td.setAttribute("nowrap","");
  text = document.eclipseform.latd.value;
  text += "\u00b0 ";
  if (document.eclipseform.latm.value < 10) text += "0";
  text += document.eclipseform.latm.value;
  text += "' ";
  if (document.eclipseform.lats.value < 10) text += "0";
  text += document.eclipseform.lats.value;
  text += '" ';
  text += (document.eclipseform.latx.options[document.eclipseform.latx.selectedIndex]).text;
  td.appendChild(document.createTextNode(text));
  row.appendChild(td);
  tbody.appendChild(row);
  row = document.createElement("tr");
  td = document.createElement("td");
  td.setAttribute("align","right");
  td.setAttribute("nowrap","");
  td.appendChild(document.createTextNode("Longitude: "));
  row.appendChild(td);
  td = document.createElement("td");
  td.setAttribute("nowrap","");
  text = document.eclipseform.lond.value;
  text += "\u00b0 ";
  if (document.eclipseform.lonm.value < 10) text += "0";
  text += document.eclipseform.lonm.value;
  text += "' ";
  if (document.eclipseform.lons.value < 10) text += "0";
  text += document.eclipseform.lons.value;
  text += '" ';
  text += (document.eclipseform.lonx.options[document.eclipseform.lonx.selectedIndex]).text;
  td.appendChild(document.createTextNode(text));
  row.appendChild(td);
  tbody.appendChild(row);
  row = document.createElement("tr");
  td = document.createElement("td");
  td.setAttribute("align","right");
  td.setAttribute("nowrap","");
  td.appendChild(document.createTextNode("Altitude: "));
  row.appendChild(td);
  td = document.createElement("td");
  td.setAttribute("nowrap","");
  text = document.eclipseform.alt.value;
  text += "m";
  td.appendChild(document.createTextNode(text));
  row.appendChild(td);
  tbody.appendChild(row);
  row = document.createElement("tr");
  td = document.createElement("td");
  td.setAttribute("align","right");
  td.setAttribute("nowrap","");
  td.appendChild(document.createTextNode("Time Zone: "));
  row.appendChild(td);
  td = document.createElement("td");
  td.setAttribute("nowrap","");
  text = (document.eclipseform.tzh.options[document.eclipseform.tzh.selectedIndex]).text;
  text += ":";
  text += (document.eclipseform.tzm.options[document.eclipseform.tzm.selectedIndex]).text;
  text += " ";
  text += (document.eclipseform.tzx.options[document.eclipseform.tzx.selectedIndex]).text;
  td.appendChild(document.createTextNode(text));
  row.appendChild(td);
  tbody.appendChild(row);
  resultsTable.appendChild(tbody);
  p.appendChild(resultsTable);
  results.appendChild(p);

  resultsTable = document.createElement("table");
  resultsTable.setAttribute("id","el_resultstable");
  resultsTable.setAttribute("width","150");
  resultsTable.setAttribute("border","2");
  tbody = document.createElement("tbody");
  row = document.createElement("tr");
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Calendar Date"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Ecl. Type"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Pen. Mag."));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Umbral Mag."));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Pen. Eclipse Begins"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Alt"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Partial Eclipse Begins"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Alt"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Total Eclipse Begins"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Alt"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Mid. Eclipse"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Alt"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Total Eclipse Ends"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Alt"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Partial Eclipse Ends"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Alt"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Pen. Eclipse Ends"));
  row.appendChild(td);
  td = document.createElement("th");
  td.appendChild(document.createTextNode("Alt"));
  row.appendChild(td);
  tbody.appendChild(row);
  for (i = 1 ; i < el.length ; i+=22) {
   if (el[5+i] <= obsvconst[5]) {
    obsvconst[4]=i;
    getall(el)
    // Is there an event...
    if (mid[5] != 1) {
      row = document.createElement("tr");
      // Calendar Date
      td = document.createElement("td");
      td.setAttribute("nowrap","")
      val = document.createTextNode(getdate(el,p1));
      td.appendChild(val);
      row.appendChild(td);
      // Eclipse Type
      td = document.createElement("td");
      td.setAttribute("align","center")
      if (el[5+i] == 1) {
        td.appendChild(document.createTextNode("T"))
      } else if (el[5+i] == 2) {
        td.appendChild(document.createTextNode("P"))
      } else {
        td.appendChild(document.createTextNode("N"))
      }
      row.appendChild(td);
      // Pen. Mag
      td = document.createElement("td");
      td.setAttribute("align","right");
      td.appendChild(document.createTextNode(el[3+i]));
      row.appendChild(td);
      // Umbral Mag
      td = document.createElement("td");
      td.setAttribute("align","right");
      td.appendChild(document.createTextNode(el[4+i]));
      row.appendChild(td);
      // P1
      td = document.createElement("td");
      td.setAttribute("align","center")
      td.appendChild(gettime(el, p1));
      row.appendChild(td)
      // P1 alt
      td = document.createElement("td");
      td.setAttribute("align","center")
      td.appendChild(getalt(p1));
      row.appendChild(td)
      if (u1[5] == 1) {
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(document.createTextNode("-"))
        row.appendChild(td);
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(document.createTextNode("-"))
        row.appendChild(td);
      } else {
        // U1
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(gettime(el, u1));
        row.appendChild(td)
        // U1 alt
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(getalt(u1));
        row.appendChild(td)
      }
      if (u2[5] == 1) {
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(document.createTextNode("-"))
        row.appendChild(td);
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(document.createTextNode("-"))
        row.appendChild(td);
      } else {
        // U2
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(gettime(el, u2));
        row.appendChild(td)
        // U2 alt
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(getalt(u2));
        row.appendChild(td)
      }
      // mid
      td = document.createElement("td");
      td.setAttribute("align","center")
      td.appendChild(gettime(el, mid));
      row.appendChild(td)
      // mid alt
      td = document.createElement("td");
      td.setAttribute("align","center")
      td.appendChild(getalt(mid));
      row.appendChild(td)
      if (u3[5] == 1) {
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(document.createTextNode("-"))
        row.appendChild(td);
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(document.createTextNode("-"))
        row.appendChild(td);
      } else {
        // u3
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(gettime(el, u3));
        row.appendChild(td)
        // u3 alt
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(getalt(u3));
        row.appendChild(td)
      }
      if (u4[5] == 1) {
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(document.createTextNode("-"))
        row.appendChild(td);
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(document.createTextNode("-"))
        row.appendChild(td);
      } else {
        // u4
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(gettime(el, u4));
        row.appendChild(td)
        // u4 alt
        td = document.createElement("td");
        td.setAttribute("align","center")
        td.appendChild(getalt(u4));
        row.appendChild(td)
      }
      // P4
      td = document.createElement("td");
      td.setAttribute("align","center")
      td.appendChild(gettime(el, p4));
      row.appendChild(td)
      // P4 alt
      td = document.createElement("td");
      td.setAttribute("align","center")
      td.appendChild(getalt(p4));
      row.appendChild(td)
      tbody.appendChild(row);
    }
   }
  }
  resultsTable.appendChild(tbody);
  results.appendChild(resultsTable);
}

function init() {
  opt = document.getElementById("el_cities");
  for (i = 0 ; i < cities.length ; i+=9) {
    name = String(cities[i]);
    if (name.indexOf("--",0) == 0) {
      group = document.createElement("optgroup");
      group.setAttribute("label",name.substring(2));
      opt.appendChild(group);
      i++;
    }
    city = document.createElement("option");
    city.setAttribute("value",i);
    cityName = document.createTextNode(cities[i]);
    city.appendChild(cityName);
    opt.appendChild(city);
  }
}

function citychange() {
  clearoldresults();
  index = Number(document.eclipseform.cityndx.value);
  if (index <= 0) return;
  hemisphere=0;
  document.eclipseform.loc_name.value = cities[index++];
  val = cities[index++];
  if (val < 0) { val=-val; hemisphere=1; }
  document.eclipseform.latd.value = val;
  val = cities[index++];
  if (val < 0) { val=-val; hemisphere=1; }
  document.eclipseform.latm.value = val;
  val = cities[index++];
  if (val < 0) { val=-val; hemisphere=1; }
  document.eclipseform.lats.value = val;
  document.eclipseform.latx.selectedIndex = hemisphere;
  hemisphere=0;
  val = cities[index++];
  if (val < 0) { val=-val; hemisphere=1; }
  document.eclipseform.lond.value = val;
  val = cities[index++];
  if (val < 0) { val=-val; hemisphere=1; }
  document.eclipseform.lonm.value = val;
  val = cities[index++];
  if (val < 0) { val=-val; hemisphere=1; }
  document.eclipseform.lons.value = val;
  document.eclipseform.lonx.selectedIndex = hemisphere;
  document.eclipseform.alt.value=cities[index++];
  val = cities[index];
  if (val < 0) {
    document.eclipseform.tzx.selectedIndex=1;
    val = -val;
  } else {
    document.eclipseform.tzx.selectedIndex=0;
  }
  document.eclipseform.tzh.selectedIndex = Math.floor(val);
  document.eclipseform.tzm.selectedIndex = Math.floor(4*(val-document.eclipseform.tzh.selectedIndex)+0.5);
}

function newloc() {
  document.eclipseform.cityndx.selectedIndex=0;
  clearoldresults();
}

function settimeperiod(timeperiod) {
  for (i = 0 ; i < loadedtimeperiods.length ; i++) {
    if (loadedtimeperiods[i] == timeperiod) {
      if (eval("self."+timeperiod)) {
        currenttimeperiod = timeperiod;
        eval(timeperiod+"()");
      }
      return;
    }
  }
  currenttimeperiod = timeperiod;
  loadedtimeperiods.push(timeperiod);
  head = document.getElementsByTagName("head")[0];
  script = document.createElement("script");
  script.setAttribute("language","JavaScript");
  script.setAttribute("src",timeperiod+".js");
  script.type = "text/javascript";
  script.defer = false;
  head.appendChild(script);
}

function recalculate() {
  eval(currenttimeperiod+"()");
}




#!/usr/bin/env python

#******************************************************************************
#
#  rpnAstronomy.py
#
#  rpnChilada astronomical operators
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import ephem

from mpmath import acos, fadd, fdiv, fmul, fsub, mpmathify, pi, power, sqrt
from pytz import timezone
from skyfield import almanac

from rpn.rpnDateTime import RPNDateTime
from rpn.rpnLocation import getLocation, getTimeZone, RPNLocation
from rpn.rpnMatchUnitTypes import matchUnitTypes
from rpn.rpnMeasurement import RPNMeasurement
from rpn.rpnMath import subtract
from rpn.rpnUtils import oneArgFunctionEvaluator, twoArgFunctionEvaluator, \
                         loadAstronomyData, validateRealInt

import rpn.rpnGlobals as g


#******************************************************************************
#
#  class RPNAstronomicalObject
#
#******************************************************************************

class RPNAstronomicalObject( ):
    '''This class is for identifying astronomical objects.'''
    def __init__( self, object ):
        self.object = object

    def getDistanceFromEarth( self, date=None ):
        if date:
            self.object.compute( date.to( 'utc' ).format( ) )

        return RPNMeasurement( fmul( self.object.earth_distance, ephem.meters_per_au ), 'meters' )

    def getAngularSize( self, location=None, date=None ):
        if location and date:
            if isinstance( location, str ):
                location = getLocation( location )

            location.observer.date = date.to( 'utc' ).format( )
            self.object.compute( location.observer )

        # I have no idea why size seems to return the value in arcseconds... that
        # goes against the pyephem documentation that it always uses radians for angles.
        return RPNMeasurement( mpmathify( fdiv( fmul( fdiv( self.object.size, 3600 ), pi ), 180 ) ), 'radian' )

    def getAngularSeparation( self, other, location, date ):
        if isinstance( location, str ):
            location = getLocation( location )

        location.observer.date = date.to( 'utc' ).format( )

        self.object.compute( location.observer )
        other.object.compute( location.observer )

        return RPNMeasurement( mpmathify( ephem.separation( ( self.object.az, self.object.alt ),
                                                            ( other.object.az, other.object.alt ) ) ), 'radian' )

    def getAzimuthAndAltitude( self, location=None, date=None ):
        if location and date:
            if isinstance( location, str ):
                location = getLocation( location )

            location.observer.date = date.to( 'utc' ).format( )
            self.object.compute( location.observer )

        return RPNMeasurement( mpmathify( self.object.az ), 'radians' ), \
               RPNMeasurement( mpmathify( self.object.alt ), 'radians' )

    def getAstronomicalEvent( self, location, date, func, horizon=None, useCenter=False, matchUSNO=False ):
        if isinstance( location, str ):
            location = getLocation( location )

        if horizon is None:
            horizon = float( location.observer.horizon )

        if matchUSNO:
            location.pressure = 0
            horizon -= 34/60        # 34 arcminutes

        oldHorizon = location.observer.horizon

        location.observer.date = date.to( 'utc' ).format( )
        location.observer.horizon = str( horizon )

        if useCenter:
            result = RPNDateTime.convertFromEphemDate( func( location.observer, self.object, use_center=useCenter ) )
        else:
            result = RPNDateTime.convertFromEphemDate( func( location.observer, self.object ) )

        location.observer.horizon = oldHorizon

        return result

    def getNextRising( self, arg1, arg2, horizon=None, useCenter=False, matchUSNO=False ):
        validUnitTypes = [ [ 'location', 'datetime' ] ]
        arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

        if not arguments:
            raise ValueError( 'unexpected arguments' )

        return self.getAstronomicalEvent( arguments[ 'location' ], arguments[ 'datetime' ],
                                          ephem.Observer.next_rising, horizon, useCenter, matchUSNO )

    def getNextSetting( self, arg1, arg2, horizon=None, useCenter=False, matchUSNO=False ):
        validUnitTypes = [ [ 'location', 'datetime' ] ]
        arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

        if not arguments:
            raise ValueError( 'unexpected arguments' )

        return self.getAstronomicalEvent( arguments[ 'location' ], arguments[ 'datetime' ],
                                          ephem.Observer.next_setting, horizon, useCenter, matchUSNO )

    def getNextTransit( self, arg1, arg2, horizon=None, useCenter=False, matchUSNO=False ):
        validUnitTypes = [ [ 'location', 'datetime' ] ]
        arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

        if not arguments:
            raise ValueError( 'unexpected arguments' )

        return self.getAstronomicalEvent( arguments[ 'location' ], arguments[ 'datetime' ],
                                          ephem.Observer.next_transit, horizon, useCenter, matchUSNO )

    def getNextAntitransit( self, arg1, arg2, horizon=None, useCenter=False, matchUSNO=False ):
        validUnitTypes = [ [ 'location', 'datetime' ] ]
        arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

        if not arguments:
            raise ValueError( 'unexpected arguments' )

        return self.getAstronomicalEvent( arguments[ 'location' ], arguments[ 'datetime' ],
                                          ephem.Observer.next_antitransit, horizon, useCenter, matchUSNO )

    def getPreviousRising( self, arg1, arg2, horizon=None, useCenter=False, matchUSNO=False ):
        validUnitTypes = [ [ 'location', 'datetime' ] ]
        arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

        if not arguments:
            raise ValueError( 'unexpected arguments' )

        return self.getAstronomicalEvent( arguments[ 'location' ], arguments[ 'datetime' ],
                                          ephem.Observer.previous_rising, horizon, useCenter, matchUSNO )

    def getPreviousSetting( self, arg1, arg2, horizon=None, useCenter=False, matchUSNO=False ):
        validUnitTypes = [ [ 'location', 'datetime' ] ]
        arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

        if not arguments:
            raise ValueError( 'unexpected arguments' )

        return self.getAstronomicalEvent( arguments[ 'location' ], arguments[ 'datetime' ],
                                          ephem.Observer.previous_setting, horizon, useCenter, matchUSNO )

    def getPreviousTransit( self, arg1, arg2, horizon=None, useCenter=False, matchUSNO=False ):
        validUnitTypes = [ [ 'location', 'datetime' ] ]
        arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

        if not arguments:
            raise ValueError( 'unexpected arguments' )

        return self.getAstronomicalEvent( arguments[ 'location' ], arguments[ 'datetime' ],
                                          ephem.Observer.previous_transit, horizon, useCenter, matchUSNO )

    def getPreviousAntitransit( self, arg1, arg2, horizon=None, useCenter=False, matchUSNO=False ):
        validUnitTypes = [ [ 'location', 'datetime' ] ]
        arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

        if not arguments:
            raise ValueError( 'unexpected arguments' )

        return self.getAstronomicalEvent( arguments[ 'location' ], arguments[ 'datetime' ],
                                          ephem.Observer.previous_antitransit, horizon, useCenter, matchUSNO )

    def getTransitTime( self, arg1, arg2, horizon=None, useCenter=False, matchUSNO=False ):
        validUnitTypes = [ [ 'location', 'datetime' ] ]
        arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

        if not arguments:
            raise ValueError( 'unexpected arguments' )

        result1 = self.getAstronomicalEvent( arguments[ 'location' ], arguments[ 'datetime' ],
                                             ephem.Observer.next_rising, horizon, useCenter, matchUSNO )
        result2 = self.getAstronomicalEvent( arguments[ 'location' ], result1,
                                             ephem.Observer.next_setting, horizon, useCenter, matchUSNO )

        return subtract( result2, result1 )

    def getAntitransitTime( self, arg1, arg2, horizon=None, useCenter=False, matchUSNO=False ):
        validUnitTypes = [ [ 'location', 'datetime' ] ]
        arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

        if not arguments:
            raise ValueError( 'unexpected arguments' )

        result1 = self.getAstronomicalEvent( arguments[ 'location' ], arguments[ 'datetime' ],
                                             ephem.Observer.next_setting, horizon, useCenter, matchUSNO )
        result2 = self.getAstronomicalEvent( arguments[ 'location' ], result1,
                                             ephem.Observer.next_rising, horizon, useCenter, matchUSNO )

        return subtract( result2, result1 )


#******************************************************************************
#
#  getSeason
#
#  0 = spring, 1 = summer, 2 = autumn, 3 = winter
#
#******************************************************************************

def getSeason( n, season ):
    '''Returns the date of the season for year n.'''
    loadAstronomyData( )

    if not g.astroDataAvailable:
        raise ValueError( 'Astronomy functions are unavailable.' )

    times, _ = almanac.find_discrete( g.timescale.utc( validateRealInt( n ), 1, 1 ),
                                      g.timescale.utc( n, 12, 31 ), almanac.seasons( g.ephemeris ) )
    result = RPNDateTime.parseDateTime( times[ season ].utc_datetime( ) )
    return result.getLocalTime( )


#******************************************************************************
#
#  getVernalEquinox
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getVernalEquinox( n ):
    '''Returns the date of the vernal equinox for year n.'''
    return getSeason( n, 0 )



#*****************************************************************************
#
#  getSummerSolstice
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getSummerSolstice( n ):
    '''Returns the date of the summer solstice for year n.'''
    return getSeason( n, 1 )


#******************************************************************************
#
#  getAutumnalEquinox
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getAutumnalEquinox( n ):
    '''Returns the date of the autumnal equinox for year n.'''
    return getSeason( n, 2 )


#*****************************************************************************
#
#  getWinterSolstice
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getWinterSolstice( n ):
    '''Returns the date of the winter solstice for year n.'''
    return getSeason( n, 3 )


#******************************************************************************
#
#  getEphemTime
#
#******************************************************************************

def getEphemTime( n, func ):
    '''Returns a pyephem date-time value from an RPNDateTime value.'''
    if not isinstance( n, RPNDateTime ):
        raise ValueError( 'expected a date-time argument' )

    result = RPNDateTime.convertFromEphemDate( func( n.format( ) ) )
    return result.getLocalTime( )

@oneArgFunctionEvaluator( )
def getNextFirstQuarterMoon( n ):
    return getEphemTime( n, ephem.next_first_quarter_moon )

@oneArgFunctionEvaluator( )
def getNextFullMoon( n ):
    return getEphemTime( n, ephem.next_full_moon )

@oneArgFunctionEvaluator( )
def getNextLastQuarterMoon( n ):
    return getEphemTime( n, ephem.next_last_quarter_moon )

@oneArgFunctionEvaluator( )
def getNextNewMoon( n ):
    return getEphemTime( n, ephem.next_new_moon )

@oneArgFunctionEvaluator( )
def getPreviousFirstQuarterMoon( n ):
    return getEphemTime( n, ephem.previous_first_quarter_moon )

@oneArgFunctionEvaluator( )
def getPreviousFullMoon( n ):
    return getEphemTime( n, ephem.previous_full_moon )

@oneArgFunctionEvaluator( )
def getPreviousLastQuarterMoon( n ):
    return getEphemTime( n, ephem.previous_last_quarter_moon )

@oneArgFunctionEvaluator( )
def getPreviousNewMoon( n ):
    return getEphemTime( n, ephem.previous_new_moon )


#******************************************************************************
#
#  getMoonPhase
#
#******************************************************************************

@oneArgFunctionEvaluator( )
def getMoonPhase( n ):
    '''Returns the current moon phase as a percentage, starting from the new moon.'''
    if not isinstance( n, RPNDateTime ):
        raise ValueError( '\'moon_phase\' expects a date-time argument' )

    datetime = n.format( )

    previous = RPNDateTime.convertFromEphemDate( ephem.previous_new_moon( datetime ) )
    next = RPNDateTime.convertFromEphemDate( ephem.next_new_moon( datetime ) )

    cycle = next - previous
    current = n - previous

    return current.total_seconds( ) / cycle.total_seconds( )


#******************************************************************************
#
#  getSkyLocation
#
#******************************************************************************

def getSkyLocation( arg1, arg2, arg3 ):
    '''Returns the location of an astronomical object in the sky in terms of azimuth and altitude.'''
    validUnitTypes = [ [ 'location', 'datetime', 'body' ] ]
    arguments = matchUnitTypes( [ arg1, arg2, arg3 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    azimuth, altitude = arguments[ 'body' ].getAzimuthAndAltitude( arguments[ 'location' ], arguments[ 'datetime' ] )

    return [ azimuth.convert( 'degree' ), altitude.convert( 'degree' ) ]


#******************************************************************************
#
#  getAngularSize
#
#******************************************************************************

def getAngularSize( arg1, arg2, arg3 ):
    '''Returns the angular size of an astronomical object in radians.'''
    validUnitTypes = [ [ 'location', 'datetime', 'body' ] ]
    arguments = matchUnitTypes( [ arg1, arg2, arg3 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    return arguments[ 'body' ].getAngularSize( arguments[ 'location' ], arguments[ 'datetime' ] )


#******************************************************************************
#
#  getAngularSeparation
#
#******************************************************************************

def getAngularSeparation( body1, body2, location, date ):
    '''Returns the angular size of an astronomical object in radians.'''
    if isinstance( location, str ):
        location = getLocation( location )

    if not isinstance( body1, RPNAstronomicalObject ) or not isinstance( body2, RPNAstronomicalObject ) and \
       not isinstance( location, RPNLocation ) or not isinstance( date, RPNDateTime ):
        raise ValueError( 'expected two astronomical objects, a location and a date-time' )

    return body1.getAngularSeparation( body2, location, date )


#******************************************************************************
#
#  getNextRising
#
#******************************************************************************

def getNextRising( arg1, arg2, arg3 ):
    '''Returns the next rising time for an astronomical object.'''
    validUnitTypes = [ [ 'location', 'datetime', 'body' ] ]
    arguments = matchUnitTypes( [ arg1, arg2, arg3 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    result = arguments[ 'body' ].getNextRising( arguments[ 'location' ], arguments[ 'datetime' ] )
    return result.getLocalTime( timezone( getTimeZone( arguments[ 'location' ] ) ) )

@twoArgFunctionEvaluator( )
def getNextSunrise( n, k ):
    return getNextRising( RPNAstronomicalObject( ephem.Sun( ) ), n, k )

@twoArgFunctionEvaluator( )
def getNextMoonRise( n, k ):
    return getNextRising( RPNAstronomicalObject( ephem.Moon( ) ), n, k )


#******************************************************************************
#
#  getNextSetting
#
#******************************************************************************

def getNextSetting( arg1, arg2, arg3 ):
    '''Returns the next setting time for an astronomical object.'''
    validUnitTypes = [ [ 'location', 'datetime', 'body' ] ]
    arguments = matchUnitTypes( [ arg1, arg2, arg3 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    result = arguments[ 'body' ].getNextSetting( arguments[ 'location' ], arguments[ 'datetime' ] )
    return result.getLocalTime( timezone( getTimeZone( arguments[ 'location' ] ) ) )

@twoArgFunctionEvaluator( )
def getNextSunset( n, k ):
    return getNextSetting( RPNAstronomicalObject( ephem.Sun( ) ), n, k )

@twoArgFunctionEvaluator( )
def getNextMoonSet( n, k ):
    return getNextSetting( RPNAstronomicalObject( ephem.Moon( ) ), n, k )


#******************************************************************************
#
#  getNextTransit
#
#******************************************************************************

def getNextTransit( arg1, arg2, arg3 ):
    validUnitTypes = [ [ 'location', 'datetime', 'body' ] ]
    arguments = matchUnitTypes( [ arg1, arg2, arg3 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    result = arguments[ 'body' ].getNextTransit( arguments[ 'location' ], arguments[ 'datetime' ] )
    return result.getLocalTime( timezone( getTimeZone( arguments[ 'location' ] ) ) )

@twoArgFunctionEvaluator( )
def getSolarNoon( n, k ):
    return getNextTransit( RPNAstronomicalObject( ephem.Sun( ) ), n, k )

@twoArgFunctionEvaluator( )
def getNextMoonTransit( n, k ):
    return getNextTransit( RPNAstronomicalObject( ephem.Moon( ) ), n, k )


#******************************************************************************
#
#  getNextAntitransit
#
#******************************************************************************

def getNextAntitransit( arg1, arg2, arg3 ):
    validUnitTypes = [ [ 'location', 'datetime', 'body' ] ]
    arguments = matchUnitTypes( [ arg1, arg2, arg3 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    result = arguments[ 'body' ].getNextAntitransit( arguments[ 'location' ], arguments[ 'datetime' ] )
    return result.getLocalTime( timezone( getTimeZone( arguments[ 'location' ] ) ) )

@twoArgFunctionEvaluator( )
def getNextSunAntitransit( n, k ):
    return getNextAntitransit( RPNAstronomicalObject( ephem.Sun( ) ), n, k )

@twoArgFunctionEvaluator( )
def getNextMoonAntitransit( n, k ):
    return getNextAntitransit( RPNAstronomicalObject( ephem.Moon( ) ), n, k )


#******************************************************************************
#
#  getTransitTime
#
#******************************************************************************

def getTransitTime( arg1, arg2, arg3 ):
    validUnitTypes = [ [ 'location', 'datetime', 'body' ] ]
    arguments = matchUnitTypes( [ arg1, arg2, arg3 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    return arguments[ 'body' ].getTransitTime( arguments[ 'location' ], arguments[ 'datetime' ] )

@twoArgFunctionEvaluator( )
def getDayTime( n, k ):
    return getTransitTime( RPNAstronomicalObject( ephem.Sun( ) ), n, k )


#******************************************************************************
#
#  getAntitransitTime
#
#******************************************************************************

def getAntitransitTime( arg1, arg2, arg3 ):
    validUnitTypes = [ [ 'location', 'datetime', 'body' ] ]
    arguments = matchUnitTypes( [ arg1, arg2, arg3 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    return arguments[ 'body' ].getAntitransitTime( arguments[ 'location' ], arguments[ 'datetime' ] )

@twoArgFunctionEvaluator( )
def getNightTime( n, k ):
    return getAntitransitTime( RPNAstronomicalObject( ephem.Sun( ) ), n, k )


#******************************************************************************
#
#  getPreviousRising
#
#******************************************************************************

def getPreviousRising( arg1, arg2, arg3 ):
    '''Returns the previous rising time for an astronomical object.'''
    validUnitTypes = [ [ 'location', 'datetime', 'body' ] ]
    arguments = matchUnitTypes( [ arg1, arg2, arg3 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    result = arguments[ 'body' ].getPreviousRising( arguments[ 'location' ], arguments[ 'datetime' ] )
    return result.getLocalTime( timezone( getTimeZone( arguments[ 'location' ] ) ) )


#******************************************************************************
#
#  getPreviousSetting
#
#******************************************************************************

def getPreviousSetting( arg1, arg2, arg3 ):
    '''Returns the previous setting time for an astronomical object.'''
    validUnitTypes = [ [ 'location', 'datetime', 'body' ] ]
    arguments = matchUnitTypes( [ arg1, arg2, arg3 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    result = arguments[ 'body' ].getPreviousSetting( arguments[ 'location' ], arguments[ 'datetime' ] )
    return result.getLocalTime( timezone( getTimeZone( arguments[ 'location' ] ) ) )


#******************************************************************************
#
#  getPreviousTransit
#
#******************************************************************************

def getPreviousTransit( arg1, arg2, arg3 ):
    validUnitTypes = [ [ 'location', 'datetime', 'body' ] ]
    arguments = matchUnitTypes( [ arg1, arg2, arg3 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    result = arguments[ 'body' ].getPreviousTransit( arguments[ 'location' ], arguments[ 'datetime' ] )
    return result.getLocalTime( timezone( getTimeZone( arguments[ 'location' ] ) ) )


#******************************************************************************
#
#  getPreviousAntitransit
#
#******************************************************************************

def getPreviousAntitransit( arg1, arg2, arg3 ):
    validUnitTypes = [ [ 'location', 'datetime', 'body' ] ]
    arguments = matchUnitTypes( [ arg1, arg2, arg3 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    result = arguments[ 'body' ].getPreviousAntitransit( arguments[ 'location' ], arguments[ 'datetime' ] )
    return result.getLocalTime( timezone( getTimeZone( arguments[ 'location' ] ) ) )


#******************************************************************************
#
#  getNextDawn
#
#  -6 is "civil" twilight
#  -12 is nautical twilight
#  -18 is astronomical twilight
#
#******************************************************************************

def getNextDawn( arg1, arg2, horizon = -6 ):
    validUnitTypes = [ [ 'location', 'datetime' ] ]
    arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    result = RPNAstronomicalObject( ephem.Sun( ) ).getNextRising( arguments[ 'location' ],
                                                                  arguments[ 'datetime' ], horizon=horizon )
    return result.getLocalTime( timezone( getTimeZone( arguments[ 'location' ] ) ) )

@twoArgFunctionEvaluator( )
def getNextCivilDawn( n, k ):
    return getNextDawn( n, k, -6 )

@twoArgFunctionEvaluator( )
def getNextNauticalDawn( n, k ):
    return getNextDawn( n, k, -12 )

@twoArgFunctionEvaluator( )
def getNextAstronomicalDawn( n, k ):
    return getNextDawn( n, k, -18 )


#******************************************************************************
#
#  getNextDusk
#
#  -6 is "civil" twilight
#  -12 is nautical twilight
#  -18 is astronomical twilight
#
#******************************************************************************

def getNextDusk( arg1, arg2, horizon = -6 ):
    validUnitTypes = [ [ 'location', 'datetime' ] ]
    arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    result = RPNAstronomicalObject( ephem.Sun( ) ).getNextSetting( arguments[ 'location' ],
                                                                   arguments[ 'datetime' ], horizon=horizon )
    return result.getLocalTime( timezone( getTimeZone( arguments[ 'location' ] ) ) )

@twoArgFunctionEvaluator( )
def getNextCivilDusk( n, k ):
    return getNextDusk( n, k, -6 )

@twoArgFunctionEvaluator( )
def getNextNauticalDusk( n, k ):
    return getNextDusk( n, k, -12 )

@twoArgFunctionEvaluator( )
def getNextAstronomicalDusk( n, k ):
    return getNextDusk( n, k, -18 )


#******************************************************************************
#
#  getDistanceFromEarth
#
#******************************************************************************

@twoArgFunctionEvaluator( )
def getDistanceFromEarth( arg1, arg2 ):
    validUnitTypes = [ [ 'body', 'datetime' ] ]
    arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

    if not arguments:
        raise ValueError( 'unexpected arguments' )

    return RPNMeasurement( arguments[ 'body' ].getDistanceFromEarth( arguments[ 'datetime' ] ), 'meters' )


#******************************************************************************
#
#  getCircleIntersectionTerm
#
#  http://mathworld.wolfram.com/Circle-CircleIntersection.html
#
#******************************************************************************

def getCircleIntersectionTerm( radius1, radius2, separation ):
    distance = fdiv( fadd( fsub( power( separation, 2 ), power( radius1, 2 ) ),
                           power( radius2, 2 ) ),
                     fmul( 2, separation ) )

    #print( 'radius1', radius1 )
    #print( 'radius2', radius2 )
    #print( 'distance', distance )
    #print( 'radius1 - distance', fsub( radius1, distance ) )
    #print( 'radius1^2 - distance^2', fsub( power( radius1, 2 ), power( distance, 2 ) ) )
    #print( )

    if power( distance, 2 ) > power( radius1, 2 ):
        return fmul( power( radius1, 2 ), fdiv( pi, 2 ) )

    return fsub( fmul( power( radius1, 2 ), acos( fdiv( distance, radius1 ) ) ),
                 fmul( distance, sqrt( fsub( power( radius1, 2 ), power( distance, 2 ) ) ) ) )


#******************************************************************************
#
#  getEclipseTotality
#
#******************************************************************************

def getEclipseTotality( body1, body2, arg1, arg2 ):
    '''Returns the angular size of an astronomical object in radians.'''
    validUnitTypes = [ [ 'location', 'datetime' ] ]
    arguments = matchUnitTypes( [ arg1, arg2 ], validUnitTypes )

    if not isinstance( body1, RPNAstronomicalObject ) or not isinstance( body2, RPNAstronomicalObject ) or \
       not arguments:
        raise ValueError( 'expected two astronomical objects, a location and a date-time' )

    separation = body1.getAngularSeparation( body2, arguments[ 'location' ], arguments[ 'datetime' ] ).value

    radius1 = body1.getAngularSize( ).value
    radius2 = body2.getAngularSize( ).value

    if separation > fadd( radius1, radius2 ):
        return 0

    distance1 = body1.getDistanceFromEarth( arguments[ 'datetime' ] )
    distance2 = body2.getDistanceFromEarth( arguments[ 'datetime' ] )

    area1 = fmul( pi, power( radius1, 2 ) )
    area2 = fmul( pi, power( radius2, 2 ) )

    # pylint: disable=arguments-out-of-order
    areaOfIntersection = fadd( getCircleIntersectionTerm( radius1, radius2, separation ),
                               getCircleIntersectionTerm( radius2, radius1, separation ) )

    if distance1 > distance2:
        result = fdiv( areaOfIntersection, area1 )
    else:
        result = fdiv( areaOfIntersection, area2 )

    if result > 1:
        return 1
    else:
        return result


#function [Az El] = RaDec2AzEl(Ra,Dec,lat,lon,time)
#% Programed by Darin C. Koblick 01/23/2010
#%--------------------------------------------------------------------------
#% External Function Call Sequence:
#% [Az El] = RaDec2AzEl(0,0,0,-104,'1992/08/20 12:14:00')
#%
#% Worked Example: pg. 262 Vallado
#%[Az El] = RaDec2AzEl(294.9891115,-20.8235624,39.007,-104.883,'1994/05/14 13:11:20.59856')
#%[210.7514  23.9036] = RaDec2AzEl(294.9891115,-20.8235624,39.007,-104.883,'1994/05/14 13:11:20.59856')
#%
#% Worked Example: http://www.stargazing.net/kepler/altaz.html
#% [Az El] = RaDec2AzEl(344.95,42.71667,52.5,-1.91667,'1997/03/14 19:00:00')
#% [311.92258 22.40100] = RaDec2AzEl(344.95,42.71667,52.5,-1.91667,'1997/03/14 19:00:00')
#%
#% [Beta,el] = RaDec2AzEl(alpha_t,delta_t,phi,lamda,'yyyy/mm/dd hh:mm:ss')
#%
#% Function Description:
#%--------------------------------------------------------------------------
#% RaDec2AzEl will take the Right Ascension and Declination in the topocentric
#% reference frame, site latitude and longitude as well as a time in GMT
#% and output the Azimuth and Elevation in the local horizon
#% reference frame.
#%
#% Inputs:                                                       Format:
#%--------------------------------------------------------------------------
#% Topocentric Right Ascension (Degrees)                         [N x 1]
#% Topocentric Declination Angle (Degrees)                       [N x 1]
#% Lat (Site Latitude in degrees -90:90 -> S(-) N(+))            [N x 1]
#% Lon (Site Longitude in degrees -180:180 W(-) E(+))            [N x 1]
#% UTC (Coordinated Universal Time YYYY/MM/DD hh:mm:ss)          [N x 1]
#%
#% Outputs:                                                      Format:
#%--------------------------------------------------------------------------
#% Local Azimuth Angle   (degrees)                               [N x 1]
#% Local Elevation Angle (degrees)                               [N x 1]
#%
#%
#% External Source References:
#% Fundamentals of Astrodynamics and Applications
#% D. Vallado, Second Edition
#% Example 3-5. Finding Local Siderial Time (pg. 192)
#% Algorithm 28: AzElToRaDec (pg. 259)
#% -------------------------------------------------------------------------

#%Example 3-5
#[yyyy mm dd HH MM SS] = datevec(datenum(time,'yyyy/mm/dd HH:MM:SS'));
#JD = juliandate(yyyy,mm,dd,HH,MM,SS);
#T_UT1 = (JD-2451545)./36525;
#ThetaGMST = 67310.54841 + (876600*3600 + 8640184.812866).*T_UT1 ...
#+ .093104.*(T_UT1.^2) - (6.2*10^-6).*(T_UT1.^3);
#ThetaGMST = mod((mod(ThetaGMST,86400*(ThetaGMST./abs(ThetaGMST)))/240),360);
#ThetaLST = ThetaGMST + lon;

#%Equation 4-11 (Define Siderial Time LHA)
#LHA = mod(ThetaLST - Ra,360);

#%Equation 4-12 (Elevation Deg)
#El = asind(sind(lat).*sind(Dec)+cosd(lat).*cosd(Dec).*cosd(LHA));

#%Equation 4-13 / 4-14 (Adaptation) (Azimuth Deg)
#%Az = mod(atand(-(sind(LHA).*cosd(Dec)./(cosd(lat).*sind(Dec) - sind(lat).*cosd(Dec).*cosd(LHA)))),360);
#Az = mod(atan2(-sind(LHA).*cosd(Dec)./cosd(El),...
#    (sind(Dec)-sind(El).*sind(lat))./(cosd(El).*cosd(lat))).*(180/pi),360);


#function jd = juliandate(year, month, day, hour, min, sec)
#YearDur = 365.25;
#for i = length(month):-1:1
#    if (month(i)<=2)
#        year(i)=year(i)-1;
#        month(i)=month(i)+12;
#    end
#end
#A = floor(YearDur*(year+4716));
#B = floor(30.6001*(month+1));
#C = 2;
#D = floor(year/100);
#E = floor(floor(year/100)*.25);
#F = day-1524.5;
#G = (hour+(min/60)+sec/3600)/24;
#jd =A+B+C-D+E+F+G;



#!/usr/bin/env python

#******************************************************************************
#
#  rpnUnitTypes.py
#
#  rpnChilada unit type declarations
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

from rpn.rpnEstimates import accelerationTable, amountOfSubstanceTable, angleTable, areaTable, \
                              capacitanceTable, catalysisTable, chargeTable, constantTable, \
                              currentTable, dataRateTable, densityTable, dynamicViscosityTable, \
                              electricalConductanceTable, electricPotentialTable, \
                              electricalResistanceTable, energyTable, forceTable, frequencyTable, \
                              illuminanceTable, inductanceTable, informationEntropyTable, \
                              jerkTable, jounceTable, lengthTable, luminanceTable, luminousFluxTable, \
                              luminousIntensityTable, magneticFieldStrengthTable, magneticFluxTable, \
                              magneticFluxDensityTable, massTable, powerTable, pressureTable, \
                              radiationDoseTable, radiationExposureTable, radioactivityTable, \
                              radiosityTable, solidAngleTable, temperatureTable, tidalForceTable, \
                              timeTable, velocityTable, volumeTable

from rpn.rpnUnitClasses import RPNUnitTypeInfo


# https://en.wikipedia.org/wiki/Radiant_exposure#SI_radiometry_units

# https://en.wikipedia.org/wiki/List_of_physical_quantities

expandedUnitTypes = {
    'abseleration'                      : ( 'length*time^3' ),
    'absement'                          : ( 'length*time' ),
    'abserk'                            : ( 'length*time^4' ),
    'absity'                            : ( 'length*time^2' ),
    'absock'                            : ( 'length*time^8' ),
    'absop'                             : ( 'length*time^7' ),
    'absorbed_dose_rate'                : ( 'length^2/time^3' ),
    'absounce'                          : ( 'length*time^5' ),
    'absrackle'                         : ( 'length*time^6' ),
    'absrop'                            : ( 'length*time^9' ),
    'acceleration'                      : ( 'length/time^2' ),
    'angular_acceleration'              : ( 'angle/time^2' ),
    'angular_velocity'                  : ( 'angle/time' ),
    'area'                              : ( 'length^2' ),
    'area_density'                      : ( 'mass/length^2' ),
    'capacitance'                       : ( 'current^2*time^4/length^2*mass' ),
    'catalysis'                         : ( 'amount_of_substance/time' ),
    'crackle'                           : ( 'length/time^5' ),
    'current_density'                   : ( 'current/length^2' ),
    'dose_equivalent'                   : ( 'length^2/time^2' ),
    'drop'                              : ( 'length/time^8' ),
    'dynamic_viscosity'                 : ( 'mass/length*time' ),
    'electrical_conductance'            : ( 'current^2*time^3/length^2*mass' ),
    'electrical_conductivity'           : ( 'current^2*time^3/length^3*mass' ),
    'electrical_resistance'             : ( 'length^3*mass/current^2*time^3' ),
    'electric_charge'                   : ( 'current*time' ),
    'electric_charge_density'           : ( 'current*time/length^3' ),
    'electric_displacement'             : ( 'current*time/length^2' ),
    'electric_potential'                : ( 'length^2*mass/current*time^3' ),
    'energy'                            : ( 'length^2*mass/time^2' ),
    'energy_density'                    : ( 'mass/length/time^2' ),
    'entropy'                           : ( 'length^2*mass/temperature*time^2' ),  # same as heat_capacity
    'force'                             : ( 'length*mass/time^2' ),
    'frequency'                         : ( '1/time' ),
    'heat_capacity'                     : ( 'length^2*mass/temperature*time^2' ),  # same as entropy
    'illuminance'                       : ( 'luminous_intensity*length^2' ),
    'inductance'                        : ( 'length^2*mass/current^2*time^2' ),
    'intensity'                         : ( 'mass/time^3' ),
    'jerk'                              : ( 'length/time^3' ),
    'jounce'                            : ( 'length/time^4' ),
    'kinemtic_viscosity'                : ( 'length^2/time' ),
    'linear_density'                    : ( 'mass/length' ),
    'lock'                              : ( 'length*time^7' ),
    'luminous_flux'                     : ( 'luminous_intensity' ),
    'magnetic_field_strength'           : ( 'current/length' ),
    'magnetic_flux'                     : ( 'length^2*mass/current*time^2' ),
    'magnetic_flux_density'             : ( 'mass/current*length*time^2' ),
    'mass_density'                      : ( 'mass/length^3' ),
    'molar_concentration'               : ( 'amount_of_substance/length^3' ),
    'molar_energy'                      : ( 'length^2*mass/amount_of_substance*time^2' ),
    'momentum'                          : ( 'length*mass/time' ),
    'moment_of_inertia'                 : ( 'length^2*mass' ),
    'permeability'                      : ( 'length*mass/current^2*time^2' ),
    'permittivity'                      : ( 'current^2*time^4/length^3*mass' ),
    'plane_angle'                       : ( 'angle' ),
    'pop'                               : ( 'length/time^6' ),
    'power'                             : ( 'length^2*mass/time^3' ),
    'pressure'                          : ( 'mass/length*time^2' ),
    'radiance'                          : ( 'mass/angle^2*time^3' ),
    'reaction_rate'                     : ( 'amount_of_substance/length^3*time' ),
    'reluctance'                        : ( 'current*time^2/length^2*mass' ),
    'solid_angle'                       : ( 'angle^2' ),
    'specific_heat_capacity'            : ( 'length^2/temperature*time^2' ),
    'specific_volume'                   : ( 'length^3/mass' ),
    'surface_tension'                   : ( 'mass/time^2' ),
    'temperature_gradient'              : ( 'temperature/length' ),
    'thermal_conductivity'              : ( 'length*mass/temperature*time^3' ),
    'tidal_force'                       : ( '1/time^2' ),    # length/time^2/length...
                                                             # lengths cancel out and that breaks my brain
    'velocity'                          : ( 'length/time' ),
    'volume'                            : ( 'length^3' ),
    'volumetric_flow'                   : ( 'length^3/time' ),
}


#******************************************************************************
#
#  basicUnitTypes
#
#  dimension:
#      length, mass, time, charge, temperature, angle, electric_potential
#      information_entropy, luminous_intensity
#
#  baseUnit:
#      the standard or customary unit of measurement for this unit type
#
#  primitiveUnit:
#      the primitive unit (or combinations of units), which only uses
#      the "primitive" types (second, meter, mole, radian, kilogram,
#      bit, ampere, kelvin, candela)
#
#  estimateTable :
#      the table of estimates, expressed in terms of the base unit,
#      used for the 'estimate' operator
#
#  Constant is not a 'real' unit type, but it is useful for it to act like
#  one because it makes for useful operators, e.g. "20 M" for "20,000,000'.
#
#  https://en.wikipedia.org/wiki/SI_electromagnetism_units
#
#
#******************************************************************************

basicUnitTypes = {
    '_null_type' : RPNUnitTypeInfo(
        'constant',
        '_null_unit',
        '_null_unit',
        None
    ),

    'acceleration' : RPNUnitTypeInfo(
        'length/time^2',
        'meter/second^2',
        'meter/second^2',
        accelerationTable
    ),

    'amount_of_substance' : RPNUnitTypeInfo(
        'amount_of_substance',
        'mole',
        'mole',
        amountOfSubstanceTable
    ),

    'angle' : RPNUnitTypeInfo(
        'angle',
        'radian',
        'radian',
        angleTable
    ),

    'area' : RPNUnitTypeInfo(
        'length^2',
        'meter^2',
        'meter^2',
        areaTable,
    ),

    'capacitance' : RPNUnitTypeInfo(
        'current^2*time^4/length^2*mass',
        'farad',
        'ampere^2*second^4/kilogram*meter^2',
        capacitanceTable,
    ),

    'catalysis' : RPNUnitTypeInfo(
        'amount_of_substance/time',
        'katal',
        'mole/second',
        catalysisTable,
    ),

    'charge' : RPNUnitTypeInfo(
        'current*time',
        'coulomb',
        'ampere*second',
        chargeTable,
    ),

    'constant' : RPNUnitTypeInfo(
        'constant',
        'unity',
        'unity',
        constantTable,
    ),

    'current' : RPNUnitTypeInfo(
        'current',
        'ampere',
        'ampere',
        currentTable,
    ),

    'data_rate' : RPNUnitTypeInfo(
        'mass*length^2/temperature*time^3',
        'bit/second',
        'bit/second',
        dataRateTable,
    ),

    'density' : RPNUnitTypeInfo(
        'mass/length^3',
        'kilogram/liter',
        'kilogram/meter^3',
        densityTable,
    ),

    'dynamic_viscosity' : RPNUnitTypeInfo(
        'mass/length*time',
        'pascal*second',
        'kilogram/meter*second',
        dynamicViscosityTable,
    ),

    'electrical_conductance' : RPNUnitTypeInfo(
        'current^2*time^3/length^2*mass',
        'siemens',
        'ampere^2*second^3/kilogram*meter^2',
        electricalConductanceTable,
    ),

    'electrical_resistance' : RPNUnitTypeInfo(
        'length^2*mass/current^2*time^3',
        'ohm',
        'kilogram*meter^2/ampere^2*second^3',
        electricalResistanceTable,
    ),

    'electric_potential' : RPNUnitTypeInfo(
        'length^2*mass/current*time^3',
        'volt',
        'kilogram*meter^2/ampere*second^3',
        electricPotentialTable,
    ),

    'energy' : RPNUnitTypeInfo(
        'length^2*mass/time^2',
        'joule',
        'kilogram*meter^2/second^2',
        energyTable,
    ),

    'force' : RPNUnitTypeInfo(
        'length*mass/time^2',
        'newton',
        'kilogram*meter/second^2',
        forceTable,
    ),

    'frequency' : RPNUnitTypeInfo(
        '1/time',
        'hertz',
        '1/second',
        frequencyTable,
    ),

    'illuminance' : RPNUnitTypeInfo(
        'angle^2*luminous_intensity/length^2',
        'lux',
        'candela*radian^2/meter^2',
        illuminanceTable,
    ),

    'inductance' : RPNUnitTypeInfo(
        'length^2*mass/current^2*time^2',
        'henry',
        'kilogram*meter^2/ampere^2*second^2',
        inductanceTable,
    ),

    'information_entropy' : RPNUnitTypeInfo(
        'length^2*mass/temperature*time^2',
        'bit',
        'kilogram*meter^2/kelvin*second^2',
        informationEntropyTable,
    ),

    'jerk' : RPNUnitTypeInfo(
        'length/time^3',
        'meter/second^3',
        'meter/second^3',
        jerkTable
    ),

    'jounce' : RPNUnitTypeInfo(
        'length/time^4',
        'meter/second^4',
        'meter/second^4',
        jounceTable
    ),

    'length' : RPNUnitTypeInfo(
        'length',
        'meter',
        'meter',
        lengthTable,
    ),

    'luminance' : RPNUnitTypeInfo(
        'luminous_intensity/length^2',
        'candela/meter^2',
        'candela/meter^2',
        luminanceTable,
    ),

    'luminous_flux' : RPNUnitTypeInfo(
        'angle^2*luminous_intensity',
        'lumen',
        'candela*radian^2',
        luminousFluxTable,
    ),

    'luminous_intensity' : RPNUnitTypeInfo(
        'luminous_intensity',
        'candela',
        'candela',
        luminousIntensityTable,
    ),

    'magnetic_field_strength' : RPNUnitTypeInfo(
        'current/length',
        'ampere/meter',
        'ampere/meter',
        magneticFieldStrengthTable,
    ),

    'magnetic_flux' : RPNUnitTypeInfo(
        'length^2*mass/current*time^2',
        'weber',
        'kilogram*meter^2/ampere*second^2',
        magneticFluxTable,
    ),

    'magnetic_flux_density' : RPNUnitTypeInfo(
        'mass/current*time^2',
        'tesla',
        'kilogram/ampere*second^2',
        magneticFluxDensityTable,
    ),

    'mass' : RPNUnitTypeInfo(
        'mass',
        'kilogram',
        'kilogram',
        massTable,
    ),

    'power' : RPNUnitTypeInfo(
        'length^2*mass/time^3',
        'watt',
        'kilogram*meter^2/second^3',
        powerTable,
    ),

    'pressure' : RPNUnitTypeInfo(
        'mass/length*time^2',
        'pascal',
        'kilogram/meter*second^2',
        pressureTable,
    ),

    'radiation_dose' : RPNUnitTypeInfo(
        'length^2/time^2',
        'sievert',
        'meter^2/second^2',
        radiationDoseTable,
    ),

    'radiation_exposure' : RPNUnitTypeInfo(
        'current*time/mass',
        'coulomb/kilogram',
        'ampere*second/kilogram',
        radiationExposureTable,
    ),

    'radiosity' : RPNUnitTypeInfo(
        'mass/time^3',
        'watt/meter^2',
        'kilograms/second^3',
        radiosityTable,
    ),

    'solid_angle' : RPNUnitTypeInfo(
        'angle^2',
        'steradian',
        'radian^2',
        solidAngleTable,
    ),

    'temperature' : RPNUnitTypeInfo(
        'temperature',
        'kelvin',
        'kelvin',
        temperatureTable,
    ),

    'tidal_force' : RPNUnitTypeInfo(
        '1/time^2',
        '1/second^2',
        '1/second^2',
        tidalForceTable,
    ),

    'time' : RPNUnitTypeInfo(
        'time',
        'second',
        'second',
        timeTable,
    ),

    'velocity' : RPNUnitTypeInfo(
        'length/time',
        'meter/second',
        'meter/second',
        velocityTable,
    ),

    'volume' : RPNUnitTypeInfo(
        'length^3',
        'liter',
        'meter^3',
        volumeTable,
    ),
}




#!/usr/bin/env python

#******************************************************************************
#
#  rpnPolynomials.py
#
#  rpnChilada polynomial operators
#  copyright (c) 2020, Rick Gutleber (rickg@his.com)
#
#  License: GNU GPL 3.0 (see <http://www.gnu.org/licenses/gpl.html> for more
#  information).
#
#******************************************************************************

import collections

from itertools import zip_longest
from mpmath import acos, arange, chop, cos, fadd, fdiv, fmul, fneg, fprod, fsub, \
                   fsum, im, libmp, mp, mpc, polyroots, polyval, power, root, sin, sqrt

from rpn.rpnGenerator import RPNGenerator


def determinant( M ):
    # pylint: disable=invalid-name
    """
    Computes the determinant of a matrix via the Schur determinant identity.
    Input: M -- square matrix; i.e., a list of lists.
    Output: A number.  If all input elements are integers, then this will also
            be an integer.

    Examples:
    > determinant([[1,2,3,4],[1,2,3,5],[1,2,4,4],[4,3,2,1]])
    5

    Adapted from https://pypi.python.org/pypi/labmath, which carries the MIT license
    but has no copyright notice.
    """
    # TODO: What is the algorithm's complexity?
    k = len( M )
    assert all( len( r ) == k for r in M )

    if k == 1:
        return M[ 0 ][ 0 ]

    if k == 2:
        return M[ 0 ][ 0 ] * M[ 1 ][ 1 ] - M[ 0 ][ 1 ] * M[ 1 ][ 0 ]

    if k == 3:
        a, b, c = M[ 0 ][ 0 ], M[ 0 ][ 1 ], M[ 0 ][ 2 ]
        d, e, f = M[ 1 ][ 0 ], M[ 1 ][ 1 ], M[ 1 ][ 2 ]
        g, h, i = M[ 2 ][ 0 ], M[ 2 ][ 1 ], M[ 2 ][ 2 ]
        return a * e * i + b * f * g + c * d * h - a * f * h - b * d * i - c * e * g

    sign = 1

    for r in range( k ):
        if M[ r ][ 0 ] != 0:
            break
    else:
        return 0

    if r != 0:
        M[ 0 ], M[ r ], sign = M[ r ], M[ 0 ], -1

    a = M[ 0 ][ 0 ]

    aD_CB = [ [ a * M[ r ][ s ] - M[ r ][ 0 ] * M[ 0 ][ s ] for s in range( 1, k ) ] for r in range( 1, k ) ]
    d = determinant( aD_CB )
    ints = isinstance( d, int ) and isinstance( a, int )
    return sign * d // a ** ( k - 2 ) if ints else sign * d / a ** ( k - 2 )


#******************************************************************************
#
#  class Polynomial
#
#  http://stackoverflow.com/questions/5413158/multiplying-polynomials-in-python
#
#******************************************************************************

class Polynomial( ):
    '''This class represents a polynomial as a list of coefficients.'''
    def __init__( self, *args ):
        '''
        Create a polynomial in one of three ways:

        p = Polynomial( poly )              # copy constructor
        p = Polynomial( [ 1, 2, 3 ... ] )   # from sequence
        p = Polynomial( 1, 2, 3 ... )       # from scalars
        '''
        super( Polynomial, self ).__init__( )

        if len( args ) == 1:
            val = args[ 0 ]

            if isinstance( val, Polynomial ):                # copy constructor
                self.coeffs = val.coeffs[ : ]
            elif isinstance( val, collections.Iterable ):    # from sequence
                self.coeffs = list( val )
            else:                                            # from single scalar
                self.coeffs = [ val + 0 ]
        else:                                                # multiple scalars
            self.coeffs = [ i + 0 for i in args ]

        self.trim( )

    def __add__( self, val ):
        'Return self + val'
        if isinstance( val, Polynomial ):                    # add Polynomial
            res = reversed( [ a + b for a, b in zip_longest( reversed( self.coeffs ),
                                                             reversed( val.coeffs ), fillvalue = 0 ) ] )
        else:                                                # add scalar
            if self.coeffs:
                res = self.coeffs[ : ]
                res[ 0 ] += val
            else:
                res = val

        return self.__class__( res )

    def __call__( self, val ):
        'Evaluate at X == val'
        res = 0
        pwr = 1

        for coeff in self.coeffs:
            res += coeff * pwr
            pwr *= val

        return res

    def __eq__( self, val ):
        'Test self == val'
        if isinstance( val, Polynomial ):
            return self.coeffs == val.coeffs
        else:
            return len( self.coeffs ) == 1 and self.coeffs[ 0 ] == val

    def __mul__( self, val ):
        'Return self * val'
        if isinstance( val, Polynomial ):
            ours = self.coeffs
            theirs = val.coeffs
            res = [ 0 ] * ( len( ours ) + len( theirs ) - 1 )

            for selfpow, selfco in enumerate( ours ):
                for valpow, valco in enumerate( theirs ):
                    res[ selfpow + valpow ] += fmul( selfco, valco )
        else:
            res = [ coeff * val for coeff in self.coeffs ]

        return self.__class__( res )

    def __neg__( self ):
        'Return -self'
        return self.__class__( [ -coeff for coeff in self.coeffs ] )

    def __pow__( self, y, z = None ):
        raise NotImplementedError( )

    def __radd__( self, val ):
        'Return val + self'
        return self + val

    def __repr__( self ):
        return '{0}({1})'.format( self.__class__.__name__, self.coeffs )

    def __rmul__( self, val ):
        'Return val * self'
        return self * val

    def __rsub__( self, val ):
        'Return val - self'
        return -self + val

    def __str__( self ):
        'Return string formatted as aX^3 + bX^2 + c^X + d'
        res = [ ]

        for exponent, coeff in enumerate( self.coeffs ):
            if coeff:
                if exponent == 0:
                    exponent = ''
                elif exponent == 1:
                    exponent = 'x'
                else:
                    exponent = 'x^' + str( exponent )

                res.append( str( coeff ) + exponent )

        if res:
            res.reverse( )
            return ' + '.join( res )
        else:
            return '0'

    def __sub__( self, val ):
        'Return self-val'
        return self.__add__( -val )

    def trim( self ):
        'Remove leading 0-coefficients'
        while self.coeffs[ 0 ] == 0:
            del self.coeffs[ 0 ]

    def getCoefficients( self ):
        return self.coeffs

    def getDiscriminant( self ):
        '''
        Computes the discriminant of a polynomial.  The input is ordered from lowest
        degree to highest so that coeffs[k] is the coefficient of the x**k term.
        Input: coeffs -- list of numbers
        Output: A number

        Examples:
        > discriminant([1,2,3,4,5])
        10800

        Adapted from https://pypi.python.org/pypi/labmath, which carries the MIT license
        but has no copyright notice.
        '''
        result = [ ]
        a = self.coeffs[ ::-1 ]
        n = len( a ) - 1

        for x in range( n - 1 ):
            result.append( [ 0 ] * x  +  a  +  [ 0 ] * ( n - 2 - x ) )

        del a[ -1 ]

        for x in range( n ):
            a[ x ] *= n - x

        for x in range( n ):
            result.append( [ 0 ] * x  +  a  +  [ 0 ] * ( n - 1 - x ) )

        return ( -1 ) ** ( n * ( n - 1 ) // 2 ) * determinant( result ) / self.coeffs[ -1 ]


#******************************************************************************
#
#  solveQuadraticPolynomial
#
#******************************************************************************

def solveQuadraticPolynomial( a, b, c ):
    # pylint: disable=invalid-name
    '''
    This function applies the quadratic formula to solve a polynomial
    with coefficients of a, b, and c.
    '''
    if a == 0:
        if b == 0:
            raise ValueError( 'invalid expression, no variable coefficients' )

        # linear equation, one root
        return [ fdiv( fneg( c ), b ) ]
    else:
        d = sqrt( fsub( power( b, 2 ), fmul( 4, fmul( a, c ) ) ) )

        x1 = fdiv( fadd( fneg( b ), d ), fmul( 2, a ) )
        x2 = fdiv( fsub( fneg( b ), d ), fmul( 2, a ) )

        return [ x1, x2 ]


#******************************************************************************
#
#  solveCubicPolynomial
#
#  Adapted from http://www.1728.org/cubic2.htm
#
#******************************************************************************

def solveCubicPolynomial( a, b, c, d ):
    # pylint: disable=invalid-name
    '''
    This function applies the cubic formula to solve a polynomial
    with coefficients of a, b, c and d.
    '''
    if mp.dps < 50:
        mp.dps = 50

    if a == 0:
        return solveQuadraticPolynomial( b, c, d )

    f = fdiv( fsub( fdiv( fmul( 3, c ), a ), fdiv( power( b, 2 ), power( a, 2 ) ) ), 3 )

    g = fdiv( fadd( fsub( fdiv( fmul( 2, power( b, 3 ) ), power( a, 3 ) ),
                          fdiv( fprod( [ 9, b, c ] ), power( a, 2 ) ) ),
                    fdiv( fmul( 27, d ), a ) ), 27 )
    h = fadd( fdiv( power( g, 2 ), 4 ), fdiv( power( f, 3 ), 27 ) )

    # all three roots are the same
    if h == 0:
        x1 = fneg( root( fdiv( d, a ), 3 ) )
        x2 = x1
        x3 = x2
    # two imaginary and one real root
    elif h > 0:
        r = fadd( fneg( fdiv( g, 2 ) ), sqrt( h ) )

        if r < 0:
            s = fneg( root( fneg( r ), 3 ) )
        else:
            s = root( r, 3 )

        t = fsub( fneg( fdiv( g, 2 ) ), sqrt( h ) )

        if t < 0:
            u = fneg( root( fneg( t ), 3 ) )
        else:
            u = root( t, 3 )

        x1 = fsub( fadd( s, u ), fdiv( b, fmul( 3, a ) ) )

        real = fsub( fdiv( fneg( fadd( s, u ) ), 2 ), fdiv( b, fmul( 3, a ) ) )
        imaginary = fdiv( fmul( fsub( s, u ), sqrt( 3 ) ), 2 )

        x2 = mpc( real, imaginary )
        x3 = mpc( real, fneg( imaginary ) )
    # all real roots
    else:
        j = sqrt( fsub( fdiv( power( g, 2 ), 4 ), h ) )
        k = acos( fneg( fdiv( g, fmul( 2, j ) ) ) )

        if j < 0:
            l = fneg( root( fneg( j ), 3 ) )
        else:
            l = root( j, 3 )

        m = cos( fdiv( k, 3 ) )
        n = fmul( sqrt( 3 ), sin( fdiv( k, 3 ) ) )
        p = fneg( fdiv( b, fmul( 3, a ) ) )

        x1 = fsub( fmul( fmul( 2, l ), cos( fdiv( k, 3 ) ) ), fdiv( b, fmul( 3, a ) ) )
        x2 = fadd( fmul( fneg( l ), fadd( m, n ) ), p )
        x3 = fadd( fmul( fneg( l ), fsub( m, n ) ), p )

    return [ chop( x1 ), chop( x2 ), chop( x3 ) ]


#******************************************************************************
#
#  solveQuarticPolynomial
#
#  Adapted from http://www.1728.org/quartic2.htm
#
#******************************************************************************

def solveQuarticPolynomial( _a, _b, _c, _d, _e ):
    # pylint: disable=invalid-name
    '''
    This function applies the quartic formula to solve a polynomial
    with coefficients of a, b, c, d, and e.
    '''
    if mp.dps < 50:
        mp.dps = 50

    # maybe it's really an order-3 polynomial
    if _a == 0:
        return solveCubicPolynomial( _b, _c, _d, _e )

    # degenerate case, just return the two real and two imaginary 4th roots of the
    # constant term divided by the 4th root of a
    elif _b == 0 and _c == 0 and _d == 0:
        e = fdiv( _e, _a )

        f = root( _a, 4 )

        x1 = fdiv( root( fneg( e ), 4 ), f )
        x2 = fdiv( fneg( root( fneg( e ), 4 ) ), f )
        x3 = fdiv( mpc( 0, root( fneg( e ), 4 ) ), f )
        x4 = fdiv( mpc( 0, fneg( root( fneg( e ), 4 ) ) ), f )

        return [ x1, x2, x3, x4 ]

    # otherwise we have a regular quartic to solve
    b = fdiv( _b, _a )
    c = fdiv( _c, _a )
    d = fdiv( _d, _a )
    e = fdiv( _e, _a )

    # we turn the equation into a cubic that we can solve
    f = fsub( c, fdiv( fmul( 3, power( b, 2 ) ), 8 ) )
    g = fsum( [ d, fdiv( power( b, 3 ), 8 ), fneg( fdiv( fmul( b, c ), 2 ) ) ] )
    h = fsum( [ e, fneg( fdiv( fmul( 3, power( b, 4 ) ), 256 ) ),
                fmul( power( b, 2 ), fdiv( c, 16 ) ), fneg( fdiv( fmul( b, d ), 4 ) ) ] )

    y1, y2, y3 = solveCubicPolynomial( 1, fdiv( f, 2 ), fdiv( fsub( power( f, 2 ), fmul( 4, h ) ), 16 ),
                                       fneg( fdiv( power( g, 2 ), 64 ) ) )

    # pick two non-zero roots, if there are two imaginary roots, use them
    if y1 == 0:
        root1 = y2
        root2 = y3
    elif y2 == 0:
        root1 = y1
        root2 = y3
    elif y3 == 0:
        root1 = y1
        root2 = y2
    elif im( y1 ) != 0:
        root1 = y1

        if im( y2 ) != 0:
            root2 = y2
        else:
            root2 = y3
    else:
        root1 = y2
        root2 = y3

    # more variables...
    p = sqrt( root1 )
    q = sqrt( root2 )
    r = fdiv( fneg( g ), fprod( [ 8, p, q ] ) )
    s = fneg( fdiv( b, 4 ) )

    # put together the 4 roots
    x1 = fsum( [ p, q, r, s ] )
    x2 = fsum( [ p, fneg( q ), fneg( r ), s ] )
    x3 = fsum( [ fneg( p ), q, fneg( r ), s ] )
    x4 = fsum( [ fneg( p ), fneg( q ), r, s ] )

    return [ chop( x1 ), chop( x2 ), chop( x3 ), chop( x4 ) ]


#******************************************************************************
#
#  addPolynomials
#
#******************************************************************************

def addPolynomials( a, b ):
    '''Adds two polynomials.'''
    lengthDiff = len( a ) - len( b )

    if lengthDiff > 0:
        for _ in range( 0, lengthDiff ):
            b.insert( 0, 0 )
    else:
        for _ in range( 0, -lengthDiff ):
            a.insert( 0, 0 )

    result = Polynomial( a )
    result += Polynomial( b )

    return result.getCoefficients( )


#******************************************************************************
#
#  multiplyPolynomials
#
#******************************************************************************

def multiplyPolynomials( a, b ):
    '''Multiplies two polynomials together.'''
    result = Polynomial( a )
    result *= Polynomial( b )

    return result.getCoefficients( )


#******************************************************************************
#
#  evaluatePolynomial
#
#******************************************************************************

def evaluatePolynomial( a, b ):
    '''Evaluates an arbitrary polynomial a for value b.'''
    if not isinstance( a, list ):
        a = [ a ]

    if isinstance( b, list ):
        return [ evaluatePolynomial( a, i ) for i in b ]
    else:
        return polyval( a, b )


#******************************************************************************
#
#  exponentiatePolynomial
#
#******************************************************************************

def exponentiatePolynomial( n, k ):
    '''Exponentiates an arbitrary polynomial by an integral power k.'''
    if not isinstance( n, list ):
        n = [ n ]

    if isinstance( k, list ):
        return [ exponentiatePolynomial( n, i ) for i in k ]
    else:
        result = n

        for _ in arange( 0, k - 1 ):
            result = multiplyPolynomials( result, n )

        return result


#******************************************************************************
#
#  solvePolynomial
#
#******************************************************************************

def solvePolynomial( args ):
    '''Uses the mpmath solve function to numerically solve an arbitrary polynomial.'''
    if isinstance( args, RPNGenerator ):
        args = list( args )
    elif not isinstance( args, list ):
        args = [ args ]

    while args[ 0 ] == 0:
        args = args[ 1 : ]

    length = len( args )

    if length == 0:
        raise ValueError( 'invalid expression, no variable coefficients' )

    if length < 2:
        raise ValueError( "'solve' requires at least an order-1 polynomial (i.e., 2 terms)" )

    nonZeroes = 0
    nonZeroIndex = 0

    for i in range( 0, length ):
        if args[ i ] != 0:
            nonZeroes += 1
            nonZeroIndex = i

    if nonZeroes == 1 and nonZeroIndex == length - 1:
        raise ValueError( 'invalid expression, no variable coefficients' )

    if nonZeroes == 1:
        return [ 0 ] * ( length - nonZeroIndex - 1 )

    try:
        result = polyroots( args )
    except libmp.libhyper.NoConvergence:
        try:
            #  Let's try again, really hard!
            result = polyroots( args, maxsteps = 2000, extraprec = 5000 )
        except libmp.libhyper.NoConvergence:
            raise ValueError( 'polynomial failed to converge' )

    return result


#******************************************************************************
#
#  multiplyPolynomialList
#
#******************************************************************************

def multiplyPolynomialList( args ):
    '''Interprets args as a list of polynomials and returns the polynomial
    product.'''
    if isinstance( args, RPNGenerator ):
        args = list( args )
    elif isinstance( args[ 0 ][ 0 ], ( list, RPNGenerator ) ):
        return [ multiplyPolynomialList( arg ) for arg in args ]
    elif not isinstance( args, list ):
        args = [ args ]

    result = None

    for arg in args:
        if isinstance( arg, RPNGenerator ):
            arg = list( arg )
        elif not isinstance( arg, list ):
            arg = [ arg ]

        if result is None:
            # arg might be a generator pylint: disable=unnecessary-comprehension
            result = Polynomial( [ i for i in arg ] )
        else:
            # arg might be a generator
            result *= Polynomial( [ i for i in arg ] )

    return result.getCoefficients( )


#******************************************************************************
#
#  sumPolynomialList
#
#******************************************************************************

def sumPolynomialList( args ):
    '''Interprets args as a list of polynomials and returns the polynomial sum.'''
    if isinstance( args, RPNGenerator ):
        args = list( args )
    elif isinstance( args[ 0 ][ 0 ], ( list, RPNGenerator ) ):
        return [ sumPolynomialList( arg ) for arg in args ]
    elif not isinstance( args, list ):
        args = [ args ]

    result = None

    for arg in args:
        if isinstance( arg, RPNGenerator ):
            arg = list( arg )
        elif not isinstance( arg, list ):
            arg = [ arg ]

        if result is None:
            # arg might be a generator
            # pylint: disable=unnecessary-comprehension
            result = Polynomial( [ i for i in arg ] )
        else:
            # arg might be a generator
            result += Polynomial( [ i for i in arg ] )

    return result.getCoefficients( )


#******************************************************************************
#
#  getPolynomialDiscriminant
#
#******************************************************************************

def getPolynomialDiscriminant( n ):
    return Polynomial( n ).getDiscriminant( )

